from __future__ import annotations

import logging
import os
import urllib.parse
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterable, Iterator, Literal, Optional, Union

import pandas as pd
from pandas.io.formats.style import Styler

from pixeltable import DataFrame, catalog, exceptions as excs, exprs, func, share
from pixeltable.catalog import Catalog, TableVersionPath
from pixeltable.catalog.insertable_table import OnErrorParameter
from pixeltable.env import Env
from pixeltable.iterators import ComponentIterator

if TYPE_CHECKING:
    import datasets  # type: ignore[import-untyped]

    RowData = list[dict[str, Any]]
    TableDataSource = Union[
        str,
        os.PathLike,
        Path,  # OS paths, filenames, URLs
        Iterator[dict[str, Any]],  # iterator producing dictionaries of values
        RowData,  # list of dictionaries
        DataFrame,  # Pixeltable DataFrame
        pd.DataFrame,  # pandas DataFrame
        'datasets.Dataset',
        'datasets.DatasetDict',  # Huggingface datasets
    ]


_logger = logging.getLogger('pixeltable')


def init() -> None:
    """Initializes the Pixeltable environment."""
    _ = Catalog.get()


def create_table(
    path_str: str,
    schema: Optional[dict[str, Any]] = None,
    *,
    source: Optional[TableDataSource] = None,
    source_format: Optional[Literal['csv', 'excel', 'parquet', 'json']] = None,
    schema_overrides: Optional[dict[str, Any]] = None,
    on_error: Literal['abort', 'ignore'] = 'abort',
    primary_key: Optional[Union[str, list[str]]] = None,
    num_retained_versions: int = 10,
    comment: str = '',
    media_validation: Literal['on_read', 'on_write'] = 'on_write',
    if_exists: Literal['error', 'ignore', 'replace', 'replace_force'] = 'error',
    extra_args: Optional[dict[str, Any]] = None,  # Additional arguments to data source provider
) -> catalog.Table:
    """Create a new base table.

    Args:
        path_str: Path to the table.
        schema: A dictionary that maps column names to column types
        source: A data source from which a table schema can be inferred and data imported
        source_format: A hint to the format of the source data
        schema_overrides: If specified, then columns in `schema_overrides` will be given the specified types
        on_error: Determines the behavior if an error occurs while evaluating a computed column or detecting an
            invalid media file (such as a corrupt image) for one of the inserted rows.

            - If `on_error='abort'`, then an exception will be raised and the rows will not be inserted.
            - If `on_error='ignore'`, then execution will continue and the rows will be inserted. Any cells
                with errors will have a `None` value for that cell, with information about the error stored in the
                corresponding `tbl.col_name.errortype` and `tbl.col_name.errormsg` fields.
        primary_key: An optional column name or list of column names to use as the primary key(s) of the
            table.
        num_retained_versions: Number of versions of the table to retain.
        comment: An optional comment; its meaning is user-defined.
        media_validation: Media validation policy for the table.

            - `'on_read'`: validate media files at query time
            - `'on_write'`: validate media files during insert/update operations
        if_exists: Directive regarding how to handle if the path already exists.
            Must be one of the following:

            - `'error'`: raise an error
            - `'ignore'`: do nothing and return the existing table handle
            - `'replace'`: if the existing table has no views, drop and replace it with a new one
            - `'replace_force'`: drop the existing table and all its views, and create a new one
        extra_args: Additional arguments to pass to the source data provider

    Returns:
        A handle to the newly created table, or to an already existing table at the path when `if_exists='ignore'`.
            Please note the schema of the existing table may not match the schema provided in the call.

    Raises:
        Error: if

            - the path is invalid, or
            - the path already exists and `if_exists='error'`, or
            - the path already exists and is not a table, or
            - an error occurs while attempting to create the table, or
            - an error occurs while attempting to import data from the source.

    Examples:
        Create a table with an int and a string column:

        >>> tbl = pxt.create_table('my_table', schema={'col1': pxt.Int, 'col2': pxt.String})

        Create a table from a select statement over an existing table `orig_table` (this will create a new table
        containing the exact contents of the query):

        >>> tbl1 = pxt.get_table('orig_table')
        ... tbl2 = pxt.create_table('new_table', tbl1.where(tbl1.col1 < 10).select(tbl1.col2))

        Create a table if does not already exist, otherwise get the existing table:

        >>> tbl = pxt.create_table('my_table', schema={'col1': pxt.Int, 'col2': pxt.String}, if_exists='ignore')

        Create a table with an int and a float column, and replace any existing table:

        >>> tbl = pxt.create_table('my_table', schema={'col1': pxt.Int, 'col2': pxt.Float}, if_exists='replace')

        Create a table from a CSV file:

        >>> tbl = pxt.create_table('my_table', source='data.csv')
    """
    from pixeltable.io.table_data_conduit import DFTableDataConduit, UnkTableDataConduit
    from pixeltable.io.utils import normalize_primary_key_parameter

    if (schema is None) == (source is None):
        raise excs.Error('Must provide either a `schema` or a `source`')

    if schema is not None and (len(schema) == 0 or not isinstance(schema, dict)):
        raise excs.Error('`schema` must be a non-empty dictionary')

    path_obj = catalog.Path(path_str)
    if_exists_ = catalog.IfExistsParam.validated(if_exists, 'if_exists')
    media_validation_ = catalog.MediaValidation.validated(media_validation, 'media_validation')
    primary_key: Optional[list[str]] = normalize_primary_key_parameter(primary_key)
    table: catalog.Table = None
    tds = None
    data_source = None
    if source is not None:
        tds = UnkTableDataConduit(source, source_format=source_format, extra_fields=extra_args)
        tds.check_source_format()
        data_source = tds.specialize()
        data_source.src_schema_overrides = schema_overrides
        data_source.src_pk = primary_key
        data_source.infer_schema()
        schema = data_source.pxt_schema
        primary_key = data_source.pxt_pk
        is_direct_df = data_source.is_direct_df()
    else:
        is_direct_df = False

    if len(schema) == 0 or not isinstance(schema, dict):
        raise excs.Error(
            'Unable to create a proper schema from supplied `source`. Please use appropriate `schema_overrides`.'
        )

    table = Catalog.get().create_table(
        path_obj,
        schema,
        data_source.pxt_df if isinstance(data_source, DFTableDataConduit) else None,
        if_exists=if_exists_,
        primary_key=primary_key,
        comment=comment,
        media_validation=media_validation_,
        num_retained_versions=num_retained_versions,
    )
    if data_source is not None and not is_direct_df:
        fail_on_exception = OnErrorParameter.fail_on_exception(on_error)
        table.insert_table_data_source(data_source=data_source, fail_on_exception=fail_on_exception)

    return table


def create_view(
    path: str,
    base: Union[catalog.Table, DataFrame],
    *,
    additional_columns: Optional[dict[str, Any]] = None,
    is_snapshot: bool = False,
    iterator: Optional[tuple[type[ComponentIterator], dict[str, Any]]] = None,
    num_retained_versions: int = 10,
    comment: str = '',
    media_validation: Literal['on_read', 'on_write'] = 'on_write',
    if_exists: Literal['error', 'ignore', 'replace', 'replace_force'] = 'error',
) -> Optional[catalog.Table]:
    """Create a view of an existing table object (which itself can be a view or a snapshot or a base table).

    Args:
        path: A name for the view; can be either a simple name such as `my_view`, or a pathname such as
            `dir1.my_view`.
        base: [`Table`][pixeltable.Table] (i.e., table or view or snapshot) or [`DataFrame`][pixeltable.DataFrame] to
            base the view on.
        additional_columns: If specified, will add these columns to the view once it is created. The format
            of the `additional_columns` parameter is identical to the format of the `schema_or_df` parameter in
            [`create_table`][pixeltable.create_table].
        is_snapshot: Whether the view is a snapshot. Setting this to `True` is equivalent to calling
            [`create_snapshot`][pixeltable.create_snapshot].
        iterator: The iterator to use for this view. If specified, then this view will be a one-to-many view of
            the base table.
        num_retained_versions: Number of versions of the view to retain.
        comment: Optional comment for the view.
        media_validation: Media validation policy for the view.

            - `'on_read'`: validate media files at query time
            - `'on_write'`: validate media files during insert/update operations
        if_exists: Directive regarding how to handle if the path already exists.
            Must be one of the following:

            - `'error'`: raise an error
            - `'ignore'`: do nothing and return the existing view handle
            - `'replace'`: if the existing view has no dependents, drop and replace it with a new one
            - `'replace_force'`: drop the existing view and all its dependents, and create a new one

    Returns:
        A handle to the [`Table`][pixeltable.Table] representing the newly created view. If the path already
            exists and `if_exists='ignore'`, returns a handle to the existing view. Please note the schema
            or the base of the existing view may not match those provided in the call.

    Raises:
        Error: if

            - the path is invalid, or
            - the path already exists and `if_exists='error'`, or
            - the path already exists and is not a view, or
            - an error occurs while attempting to create the view.

    Examples:
        Create a view `my_view` of an existing table `my_table`, filtering on rows where `col1` is greater than 10:

        >>> tbl = pxt.get_table('my_table')
        ... view = pxt.create_view('my_view', tbl.where(tbl.col1 > 10))

        Create a view `my_view` of an existing table `my_table`, filtering on rows where `col1` is greater than 10,
        and if it not already exist. Otherwise, get the existing view named `my_view`:

        >>> tbl = pxt.get_table('my_table')
        ... view = pxt.create_view('my_view', tbl.where(tbl.col1 > 10), if_exists='ignore')

        Create a view `my_view` of an existing table `my_table`, filtering on rows where `col1` is greater than 100,
        and replace any existing view named `my_view`:

        >>> tbl = pxt.get_table('my_table')
        ... view = pxt.create_view('my_view', tbl.where(tbl.col1 > 100), if_exists='replace_force')
    """
    tbl_version_path: TableVersionPath
    select_list: Optional[list[tuple[exprs.Expr, Optional[str]]]] = None
    where: Optional[exprs.Expr] = None
    if isinstance(base, catalog.Table):
        tbl_version_path = base._tbl_version_path
    elif isinstance(base, DataFrame):
        base._validate_mutable('create_view', allow_select=True)
        if len(base._from_clause.tbls) > 1:
            raise excs.Error('Cannot create a view of a join')
        tbl_version_path = base._from_clause.tbls[0]
        where = base.where_clause
        select_list = base.select_list
    else:
        raise excs.Error('`base` must be an instance of `Table` or `DataFrame`')
    assert isinstance(base, (catalog.Table, DataFrame))

    path_obj = catalog.Path(path)
    if_exists_ = catalog.IfExistsParam.validated(if_exists, 'if_exists')
    media_validation_ = catalog.MediaValidation.validated(media_validation, 'media_validation')

    if additional_columns is None:
        additional_columns = {}
    else:
        # additional columns should not be in the base table
        for col_name in additional_columns:
            if col_name in [c.name for c in tbl_version_path.columns()]:
                raise excs.Error(
                    f'Column {col_name!r} already exists in the base table '
                    f'{tbl_version_path.get_column(col_name).tbl.get().name}.'
                )

    return Catalog.get().create_view(
        path_obj,
        tbl_version_path,
        select_list=select_list,
        where=where,
        additional_columns=additional_columns,
        is_snapshot=is_snapshot,
        iterator=iterator,
        num_retained_versions=num_retained_versions,
        comment=comment,
        media_validation=media_validation_,
        if_exists=if_exists_,
    )


def create_snapshot(
    path_str: str,
    base: Union[catalog.Table, DataFrame],
    *,
    additional_columns: Optional[dict[str, Any]] = None,
    iterator: Optional[tuple[type[ComponentIterator], dict[str, Any]]] = None,
    num_retained_versions: int = 10,
    comment: str = '',
    media_validation: Literal['on_read', 'on_write'] = 'on_write',
    if_exists: Literal['error', 'ignore', 'replace', 'replace_force'] = 'error',
) -> Optional[catalog.Table]:
    """Create a snapshot of an existing table object (which itself can be a view or a snapshot or a base table).

    Args:
        path_str: A name for the snapshot; can be either a simple name such as `my_snapshot`, or a pathname such as
            `dir1.my_snapshot`.
        base: [`Table`][pixeltable.Table] (i.e., table or view or snapshot) or [`DataFrame`][pixeltable.DataFrame] to
            base the snapshot on.
        additional_columns: If specified, will add these columns to the snapshot once it is created. The format
            of the `additional_columns` parameter is identical to the format of the `schema_or_df` parameter in
            [`create_table`][pixeltable.create_table].
        iterator: The iterator to use for this snapshot. If specified, then this snapshot will be a one-to-many view of
            the base table.
        num_retained_versions: Number of versions of the view to retain.
        comment: Optional comment for the snapshot.
        media_validation: Media validation policy for the snapshot.

            - `'on_read'`: validate media files at query time
            - `'on_write'`: validate media files during insert/update operations
        if_exists: Directive regarding how to handle if the path already exists.
            Must be one of the following:

            - `'error'`: raise an error
            - `'ignore'`: do nothing and return the existing snapshot handle
            - `'replace'`: if the existing snapshot has no dependents, drop and replace it with a new one
            - `'replace_force'`: drop the existing snapshot and all its dependents, and create a new one

    Returns:
        A handle to the [`Table`][pixeltable.Table] representing the newly created snapshot.
            Please note the schema or base of the existing snapshot may not match those provided in the call.

    Raises:
        Error: if

            - the path is invalid, or
            - the path already exists and `if_exists='error'`, or
            - the path already exists and is not a snapshot, or
            - an error occurs while attempting to create the snapshot.

    Examples:
        Create a snapshot `my_snapshot` of a table `my_table`:

        >>> tbl = pxt.get_table('my_table')
        ... snapshot = pxt.create_snapshot('my_snapshot', tbl)

        Create a snapshot `my_snapshot` of a view `my_view` with additional int column `col3`,
        if `my_snapshot` does not already exist:

        >>> view = pxt.get_table('my_view')
        ... snapshot = pxt.create_snapshot(
        ...     'my_snapshot', view, additional_columns={'col3': pxt.Int}, if_exists='ignore'
        ... )

        Create a snapshot `my_snapshot` on a table `my_table`, and replace any existing snapshot named `my_snapshot`:

        >>> tbl = pxt.get_table('my_table')
        ... snapshot = pxt.create_snapshot('my_snapshot', tbl, if_exists='replace_force')
    """
    return create_view(
        path_str,
        base,
        additional_columns=additional_columns,
        iterator=iterator,
        is_snapshot=True,
        num_retained_versions=num_retained_versions,
        comment=comment,
        media_validation=media_validation,
        if_exists=if_exists,
    )


def get_table(path: str) -> catalog.Table:
    """Get a handle to an existing table, view, or snapshot.

    Args:
        path: Path to the table.

    Returns:
        A handle to the [`Table`][pixeltable.Table].

    Raises:
        Error: If the path does not exist or does not designate a table object.

    Examples:
        Get handle for a table in the top-level directory:

        >>> tbl = pxt.get_table('my_table')

        For a table in a subdirectory:

        >>> tbl = pxt.get_table('subdir.my_table')

        Handles to views and snapshots are retrieved in the same way:

        >>> tbl = pxt.get_table('my_snapshot')
    """
    path_obj = catalog.Path(path)
    return Catalog.get().get_table(path_obj)


def move(path: str, new_path: str) -> None:
    """Move a schema object to a new directory and/or rename a schema object.

    Args:
        path: absolute path to the existing schema object.
        new_path: absolute new path for the schema object.

    Raises:
        Error: If path does not exist or new_path already exists.

    Examples:
        Move a table to a different directory:

        >>>> pxt.move('dir1.my_table', 'dir2.my_table')

        Rename a table:

        >>>> pxt.move('dir1.my_table', 'dir1.new_name')
    """
    if path == new_path:
        raise excs.Error('move(): source and destination cannot be identical')
    path_obj, new_path_obj = catalog.Path(path), catalog.Path(new_path)
    if path_obj.is_ancestor(new_path_obj):
        raise excs.Error(f'move(): cannot move {path!r} into its own subdirectory')
    cat = Catalog.get()
    cat.move(path_obj, new_path_obj)


def drop_table(
    table: Union[str, catalog.Table], force: bool = False, if_not_exists: Literal['error', 'ignore'] = 'error'
) -> None:
    """Drop a table, view, or snapshot.

    Args:
        table: Fully qualified name, or handle, of the table to be dropped.
        force: If `True`, will also drop all views and sub-views of this table.
        if_not_exists: Directive regarding how to handle if the path does not exist.
            Must be one of the following:

            - `'error'`: raise an error
            - `'ignore'`: do nothing and return

    Raises:
        Error: if the qualified name

            - is invalid, or
            - does not exist and `if_not_exists='error'`, or
            - does not designate a table object, or
            - designates a table object but has dependents and `force=False`.

    Examples:
        Drop a table by its fully qualified name:
        >>> pxt.drop_table('subdir.my_table')

        Drop a table by its handle:
        >>> t = pxt.get_table('subdir.my_table')
        ... pxt.drop_table(t)

        Drop a table if it exists, otherwise do nothing:
        >>> pxt.drop_table('subdir.my_table', if_not_exists='ignore')

        Drop a table and all its dependents:
        >>> pxt.drop_table('subdir.my_table', force=True)
    """
    tbl_path: str
    if isinstance(table, catalog.Table):
        # if we're dropping a table by handle, we first need to get the current path, then drop the S lock on
        # the Table record, and then get X locks in the correct order (first containing directory, then table)
        with Env.get().begin_xact():
            tbl_path = table._path()
    else:
        assert isinstance(table, str)
        tbl_path = table

    path_obj = catalog.Path(tbl_path)
    if_not_exists_ = catalog.IfNotExistsParam.validated(if_not_exists, 'if_not_exists')
    Catalog.get().drop_table(path_obj, force=force, if_not_exists=if_not_exists_)


def list_tables(dir_path: str = '', recursive: bool = True) -> list[str]:
    """List the [`Table`][pixeltable.Table]s in a directory.

    Args:
        dir_path: Path to the directory. Defaults to the root directory.
        recursive: If `False`, returns only those tables that are directly contained in specified directory; if
            `True`, returns all tables that are descendants of the specified directory, recursively.

    Returns:
        A list of [`Table`][pixeltable.Table] paths.

    Raises:
        Error: If the path does not exist or does not designate a directory.

    Examples:
        List tables in top-level directory:

        >>> pxt.list_tables()

        List tables in 'dir1':

        >>> pxt.list_tables('dir1')
    """
    path_obj = catalog.Path(dir_path, empty_is_valid=True)  # validate format
    cat = Catalog.get()
    contents = cat.get_dir_contents(path_obj, recursive=recursive)
    return [str(p) for p in _extract_paths(contents, parent=path_obj, entry_type=catalog.Table)]


def create_dir(
    path: str, if_exists: Literal['error', 'ignore', 'replace', 'replace_force'] = 'error', parents: bool = False
) -> Optional[catalog.Dir]:
    """Create a directory.

    Args:
        path: Path to the directory.
        if_exists: Directive regarding how to handle if the path already exists.
            Must be one of the following:

            - `'error'`: raise an error
            - `'ignore'`: do nothing and return the existing directory handle
            - `'replace'`: if the existing directory is empty, drop it and create a new one
            - `'replace_force'`: drop the existing directory and all its children, and create a new one
        parents: Create missing parent directories.

    Returns:
        A handle to the newly created directory, or to an already existing directory at the path when
            `if_exists='ignore'`. Please note the existing directory may not be empty.

    Raises:
        Error: If

            - the path is invalid, or
            - the path already exists and `if_exists='error'`, or
            - the path already exists and is not a directory, or
            - an error occurs while attempting to create the directory.

    Examples:
        >>> pxt.create_dir('my_dir')

        Create a subdirectory:

        >>> pxt.create_dir('my_dir.sub_dir')

        Create a subdirectory only if it does not already exist, otherwise do nothing:

        >>> pxt.create_dir('my_dir.sub_dir', if_exists='ignore')

        Create a directory and replace if it already exists:

        >>> pxt.create_dir('my_dir', if_exists='replace_force')

        Create a subdirectory along with its ancestors:

        >>> pxt.create_dir('parent1.parent2.sub_dir', parents=True)
    """
    path_obj = catalog.Path(path)
    if_exists_ = catalog.IfExistsParam.validated(if_exists, 'if_exists')
    return Catalog.get().create_dir(path_obj, if_exists=if_exists_, parents=parents)


def drop_dir(path: str, force: bool = False, if_not_exists: Literal['error', 'ignore'] = 'error') -> None:
    """Remove a directory.

    Args:
        path: Name or path of the directory.
        force: If `True`, will also drop all tables and subdirectories of this directory, recursively, along
            with any views or snapshots that depend on any of the dropped tables.
        if_not_exists: Directive regarding how to handle if the path does not exist.
            Must be one of the following:

            - `'error'`: raise an error
            - `'ignore'`: do nothing and return

    Raises:
        Error: If the path

            - is invalid, or
            - does not exist and `if_not_exists='error'`, or
            - is not designate a directory, or
            - is a direcotory but is not empty and `force=False`.

    Examples:
        Remove a directory, if it exists and is empty:
        >>> pxt.drop_dir('my_dir')

        Remove a subdirectory:

        >>> pxt.drop_dir('my_dir.sub_dir')

        Remove an existing directory if it is empty, but do nothing if it does not exist:

        >>> pxt.drop_dir('my_dir.sub_dir', if_not_exists='ignore')

        Remove an existing directory and all its contents:

        >>> pxt.drop_dir('my_dir', force=True)
    """
    path_obj = catalog.Path(path)  # validate format
    if_not_exists_ = catalog.IfNotExistsParam.validated(if_not_exists, 'if_not_exists')
    Catalog.get().drop_dir(path_obj, if_not_exists=if_not_exists_, force=force)


def _extract_paths(
    dir_entries: dict[str, Catalog.DirEntry],
    parent: catalog.Path,
    entry_type: Optional[type[catalog.SchemaObject]] = None,
) -> list[catalog.Path]:
    """Convert nested dir_entries structure to a flattened list of paths."""
    matches: list[str]
    if entry_type is None:
        matches = list(dir_entries.keys())
    elif entry_type is catalog.Dir:
        matches = [name for name, entry in dir_entries.items() if entry.dir is not None]
    else:
        matches = [name for name, entry in dir_entries.items() if entry.table is not None]
    result = [parent.append(name) for name in matches]
    for name, entry in [(name, entry) for name, entry in dir_entries.items() if len(entry.dir_entries) > 0]:
        result.extend(_extract_paths(entry.dir_entries, parent=parent.append(name), entry_type=entry_type))
    return result


def publish_snapshot(dest_uri: str, table: catalog.Table) -> None:
    parsed_uri = urllib.parse.urlparse(dest_uri)
    if parsed_uri.scheme != 'pxt':
        raise excs.Error(f'Invalid Pixeltable URI (does not start with pxt://): {dest_uri}')
    share.publish_snapshot(dest_uri, table)


def list_dirs(path: str = '', recursive: bool = True) -> list[str]:
    """List the directories in a directory.

    Args:
        path: Name or path of the directory.
        recursive: If `True`, lists all descendants of this directory recursively.

    Returns:
        List of directory paths.

    Raises:
        Error: If `path_str` does not exist or does not designate a directory.

    Examples:
        >>> cl.list_dirs('my_dir', recursive=True)
        ['my_dir', 'my_dir.sub_dir1']
    """
    path_obj = catalog.Path(path, empty_is_valid=True)  # validate format
    cat = Catalog.get()
    contents = cat.get_dir_contents(path_obj, recursive=recursive)
    return [str(p) for p in _extract_paths(contents, parent=path_obj, entry_type=catalog.Dir)]


def list_functions() -> Styler:
    """Returns information about all registered functions.

    Returns:
        Pandas DataFrame with columns 'Path', 'Name', 'Parameters', 'Return Type', 'Is Agg', 'Library'
    """
    functions = func.FunctionRegistry.get().list_functions()
    paths = ['.'.join(f.self_path.split('.')[:-1]) for f in functions]
    names = [f.name for f in functions]
    params = [
        ', '.join(
            [param_name + ': ' + str(param_type) for param_name, param_type in f.signatures[0].parameters.items()]
        )
        for f in functions
    ]
    pd_df = pd.DataFrame(
        {
            'Path': paths,
            'Function Name': names,
            'Parameters': params,
            'Return Type': [str(f.signatures[0].get_return_type()) for f in functions],
        }
    )
    pd_df = pd_df.style.set_properties(None, **{'text-align': 'left'}).set_table_styles(
        [{'selector': 'th', 'props': [('text-align', 'center')]}]
    )  # center-align headings
    return pd_df.hide(axis='index')


def tools(*args: Union[func.Function, func.tools.Tool]) -> func.tools.Tools:
    """
    Specifies a collection of UDFs to be used as LLM tools. Pixeltable allows any UDF to be used as an input into an
    LLM tool-calling API. To use one or more UDFs as tools, wrap them in a `pxt.tools` call and pass the return value
    to an LLM API.

    The UDFs can be specified directly or wrapped inside a [pxt.tool()][pixeltable.tool] invocation. If a UDF is
    specified directly, the tool name will be the (unqualified) UDF name, and the tool description will consist of the
    entire contents of the UDF docstring. If a UDF is wrapped in a `pxt.tool()` invocation, then the name and/or
    description may be customized.

    Args:
        args: The UDFs to use as tools.

    Returns:
        A `Tools` instance that can be passed to an LLM tool-calling API or invoked to generate tool results.

    Examples:
        Create a tools instance with a single UDF:

        >>> tools = pxt.tools(stock_price)

        Create a tools instance with several UDFs:

        >>> tools = pxt.tools(stock_price, weather_quote)

        Create a tools instance, some of whose UDFs have customized metadata:

        >>> tools = pxt.tools(
        ...     stock_price,
        ...     pxt.tool(weather_quote, description='Returns information about the weather in a particular location.'),
        ...     pxt.tool(traffic_quote, name='traffic_conditions'),
        ... )
    """
    return func.tools.Tools(tools=[arg if isinstance(arg, func.tools.Tool) else tool(arg) for arg in args])


def tool(fn: func.Function, name: Optional[str] = None, description: Optional[str] = None) -> func.tools.Tool:
    """
    Specifies a Pixeltable UDF to be used as an LLM tool with customizable metadata. See the documentation for
    [pxt.tools()][pixeltable.tools] for more details.

    Args:
        fn: The UDF to use as a tool.
        name: The name of the tool. If not specified, then the unqualified name of the UDF will be used by default.
        description: The description of the tool. If not specified, then the entire contents of the UDF docstring
            will be used by default.

    Returns:
        A `Tool` instance that can be passed to an LLM tool-calling API.
    """
    if isinstance(fn, func.AggregateFunction):
        raise excs.Error('Aggregator UDFs cannot be used as tools')

    return func.tools.Tool(fn=fn, name=name, description=description)


def configure_logging(
    *,
    to_stdout: Optional[bool] = None,
    level: Optional[int] = None,
    add: Optional[str] = None,
    remove: Optional[str] = None,
) -> None:
    """Configure logging.

    Args:
        to_stdout: if True, also log to stdout
        level: default log level
        add: comma-separated list of 'module name:log level' pairs; ex.: add='video:10'
        remove: comma-separated list of module names
    """
    return Env.get().configure_logging(to_stdout=to_stdout, level=level, add=add, remove=remove)


def array(elements: Iterable) -> exprs.Expr:
    return exprs.Expr.from_array(elements)



from __future__ import annotations

from typing import Any, Optional

import sqlalchemy as sql

import pixeltable.exceptions as excs
import pixeltable.type_system as ts

from .column_ref import ColumnRef
from .data_row import DataRow
from .expr import Expr
from .globals import ComparisonOperator
from .literal import Literal
from .row_builder import RowBuilder
from .sql_element_cache import SqlElementCache


class Comparison(Expr):
    is_search_arg_comparison: bool
    operator: ComparisonOperator

    def __init__(self, operator: ComparisonOperator, op1: Expr, op2: Expr):
        from pixeltable import index

        super().__init__(ts.BoolType())
        self.operator = operator

        # if this is a comparison of a column to a literal (ie, could be used as a search argument in an index lookup),
        # normalize it to <column> <operator> <literal>.
        if isinstance(op1, ColumnRef) and isinstance(op2, Literal):
            self.is_search_arg_comparison = True
            self.components = [op1, op2]
        elif isinstance(op1, Literal) and isinstance(op2, ColumnRef):
            self.is_search_arg_comparison = True
            self.components = [op2, op1]
            self.operator = self.operator.reverse()
        else:
            self.is_search_arg_comparison = False
            self.components = [op1, op2]

        if (
            self.is_search_arg_comparison
            and self._op2.col_type.is_string_type()
            and len(self._op2.val) >= index.BtreeIndex.MAX_STRING_LEN
        ):
            # we can't use an index for this after all
            raise excs.Error(
                f'String literal too long for comparison against indexed column {self._op1.col.name!r} '
                f'(max length is {index.BtreeIndex.MAX_STRING_LEN - 1})'
            )

        self.id = self._create_id()

    def __repr__(self) -> str:
        return f'{self._op1} {self.operator} {self._op2}'

    def _equals(self, other: Comparison) -> bool:
        return self.operator == other.operator

    def _id_attrs(self) -> list[tuple[str, Any]]:
        return [*super()._id_attrs(), ('operator', self.operator.value)]

    @property
    def _op1(self) -> Expr:
        return self.components[0]

    @property
    def _op2(self) -> Expr:
        return self.components[1]

    def sql_expr(self, sql_elements: SqlElementCache) -> Optional[sql.ColumnElement]:
        from pixeltable import index

        if str(self._op1.col_type.to_sa_type()) != str(self._op2.col_type.to_sa_type()):
            # Comparing columns of different SQL types (e.g., string vs. json); this can only be done in Python
            # TODO(aaron-siegel): We may be able to handle some cases in SQL by casting one side to the other's type
            return None

        left = sql_elements.get(self._op1)
        if self.is_search_arg_comparison:
            # reference the index value column if there is an index and this is not a snapshot
            # (indices don't apply to snapshots)
            tbl = self._op1.col.tbl.get()
            idx_info = [
                info for info in self._op1.col.get_idx_info().values() if isinstance(info.idx, index.BtreeIndex)
            ]
            if len(idx_info) > 0 and not tbl.is_snapshot:
                # there shouldn't be multiple B-tree indices on a column
                assert len(idx_info) == 1
                left = idx_info[0].val_col.sa_col

        right = sql_elements.get(self._op2)
        if left is None or right is None:
            return None

        if self.operator == ComparisonOperator.LT:
            return left < right
        if self.operator == ComparisonOperator.LE:
            return left <= right
        if self.operator == ComparisonOperator.EQ:
            return left == right
        if self.operator == ComparisonOperator.NE:
            return left != right
        if self.operator == ComparisonOperator.GT:
            return left > right
        if self.operator == ComparisonOperator.GE:
            return left >= right

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        left = data_row[self._op1.slot_idx]
        right = data_row[self._op2.slot_idx]

        if self.operator == ComparisonOperator.LT:
            data_row[self.slot_idx] = left < right
        elif self.operator == ComparisonOperator.LE:
            data_row[self.slot_idx] = left <= right
        elif self.operator == ComparisonOperator.EQ:
            data_row[self.slot_idx] = left == right
        elif self.operator == ComparisonOperator.NE:
            data_row[self.slot_idx] = left != right
        elif self.operator == ComparisonOperator.GT:
            data_row[self.slot_idx] = left > right
        elif self.operator == ComparisonOperator.GE:
            data_row[self.slot_idx] = left >= right

    def _as_dict(self) -> dict:
        return {'operator': self.operator.value, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: dict, components: list[Expr]) -> Comparison:
        assert 'operator' in d
        return cls(ComparisonOperator(d['operator']), components[0], components[1])



# ruff: noqa: F401

from .aggregate_function import AggregateFunction, Aggregator, uda
from .callable_function import CallableFunction
from .expr_template_function import ExprTemplateFunction
from .function import Function, InvalidFunction
from .function_registry import FunctionRegistry
from .query_template_function import QueryTemplateFunction, query
from .signature import Batch, Parameter, Signature
from .tools import Tool, ToolChoice, Tools
from .udf import expr_udf, make_function, udf



from __future__ import annotations

import logging
import os
import shutil
from pathlib import Path
from typing import Any, ClassVar, Optional, TypeVar

import toml

from pixeltable import exceptions as excs

_logger = logging.getLogger('pixeltable')

T = TypeVar('T')


class Config:
    """
    The (global) Pixeltable configuration, as loaded from `config.toml`. Provides methods for retrieving
    configuration values, which can be set in the config file or as environment variables.
    """

    __instance: ClassVar[Optional[Config]] = None

    __home: Path
    __config_file: Path
    __config_dict: dict[str, Any]

    def __init__(self) -> None:
        assert self.__instance is None, 'Config is a singleton; use Config.get() to access the instance'

        self.__home = Path(os.environ.get('PIXELTABLE_HOME', str(Path.home() / '.pixeltable')))
        if self.__home.exists() and not self.__home.is_dir():
            raise RuntimeError(f'{self.__home} is not a directory')
        if not self.__home.exists():
            print(f'Creating a Pixeltable instance at: {self.__home}')
            self.__home.mkdir()

        self.__config_file = Path(os.environ.get('PIXELTABLE_CONFIG', str(self.__home / 'config.toml')))

        self.__config_dict: dict[str, Any]
        if os.path.isfile(self.__config_file):
            with open(self.__config_file, 'r', encoding='utf-8') as stream:
                try:
                    self.__config_dict = toml.load(stream)
                except Exception as exc:
                    raise excs.Error(f'Could not read config file: {self.__config_file}') from exc
        else:
            self.__config_dict = self.__create_default_config(self.__config_file)
            with open(self.__config_file, 'w', encoding='utf-8') as stream:
                try:
                    toml.dump(self.__config_dict, stream)
                except Exception as exc:
                    raise excs.Error(f'Could not write config file: {self.__config_file}') from exc
            _logger.info(f'Created default config file at: {self.__config_file}')

    @property
    def home(self) -> Path:
        return self.__home

    @property
    def config_file(self) -> Path:
        return self.__config_file

    @classmethod
    def get(cls) -> Config:
        if cls.__instance is None:
            cls.__instance = cls()
        return cls.__instance

    @classmethod
    def __create_default_config(cls, config_path: Path) -> dict[str, Any]:
        free_disk_space_bytes = shutil.disk_usage(config_path.parent).free
        # Default cache size is 1/5 of free disk space
        file_cache_size_g = free_disk_space_bytes / 5 / (1 << 30)
        return {'pixeltable': {'file_cache_size_g': round(file_cache_size_g, 1), 'hide_warnings': False}}

    def get_value(self, key: str, expected_type: type[T], section: str = 'pixeltable') -> Optional[T]:
        env_var = f'{section.upper()}_{key.upper()}'
        if env_var in os.environ:
            value = os.environ[env_var]
        elif section in self.__config_dict and key in self.__config_dict[section]:
            value = self.__config_dict[section][key]
        else:
            return None

        try:
            return expected_type(value)  # type: ignore[call-arg]
        except ValueError as exc:
            raise excs.Error(f'Invalid value for configuration parameter {section}.{key}: {value}') from exc

    def get_string_value(self, key: str, section: str = 'pixeltable') -> Optional[str]:
        return self.get_value(key, str, section)

    def get_int_value(self, key: str, section: str = 'pixeltable') -> Optional[int]:
        return self.get_value(key, int, section)

    def get_float_value(self, key: str, section: str = 'pixeltable') -> Optional[float]:
        return self.get_value(key, float, section)

    def get_bool_value(self, key: str, section: str = 'pixeltable') -> Optional[bool]:
        return self.get_value(key, bool, section)



from __future__ import annotations

import enum
from typing import Any, Optional

import sqlalchemy as sql

import pixeltable.type_system as ts
from pixeltable import catalog

from .column_ref import ColumnRef
from .data_row import DataRow
from .expr import Expr
from .row_builder import RowBuilder
from .sql_element_cache import SqlElementCache


class ColumnPropertyRef(Expr):
    """A reference to a property of a table column

    The properties themselves are type-specific and may or may not need to reference the underlying column data.
    """

    class Property(enum.Enum):
        ERRORTYPE = 0
        ERRORMSG = 1
        FILEURL = 2
        LOCALPATH = 3

    def __init__(self, col_ref: ColumnRef, prop: Property):
        super().__init__(ts.StringType(nullable=True))
        self.components = [col_ref]
        self.prop = prop
        self.id = self._create_id()

    def default_column_name(self) -> Optional[str]:
        return str(self).replace('.', '_')

    def _equals(self, other: ColumnPropertyRef) -> bool:
        return self.prop == other.prop

    def _id_attrs(self) -> list[tuple[str, Any]]:
        return [*super()._id_attrs(), ('prop', self.prop.value)]

    @property
    def _col_ref(self) -> ColumnRef:
        col_ref = self.components[0]
        assert isinstance(col_ref, ColumnRef)
        return col_ref

    def __repr__(self) -> str:
        return f'{self._col_ref}.{self.prop.name.lower()}'

    def is_error_prop(self) -> bool:
        return self.prop in (self.Property.ERRORTYPE, self.Property.ERRORMSG)

    def sql_expr(self, sql_elements: SqlElementCache) -> Optional[sql.ColumnElement]:
        if not self._col_ref.col.is_stored:
            return None

        # the errortype/-msg properties of a read-validated media column need to be extracted from the DataRow
        if (
            self._col_ref.col.col_type.is_media_type()
            and self._col_ref.col.media_validation == catalog.MediaValidation.ON_READ
            and self.is_error_prop()
        ):
            return None

        if self.prop == self.Property.ERRORTYPE:
            assert self._col_ref.col.sa_errortype_col is not None
            return self._col_ref.col.sa_errortype_col
        if self.prop == self.Property.ERRORMSG:
            assert self._col_ref.col.sa_errormsg_col is not None
            return self._col_ref.col.sa_errormsg_col
        if self.prop == self.Property.FILEURL:
            # the file url is stored as the column value
            return sql_elements.get(self._col_ref)
        return None

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        if self.prop == self.Property.FILEURL:
            assert data_row.has_val[self._col_ref.slot_idx]
            data_row[self.slot_idx] = data_row.file_urls[self._col_ref.slot_idx]
            return
        elif self.prop == self.Property.LOCALPATH:
            assert data_row.has_val[self._col_ref.slot_idx]
            data_row[self.slot_idx] = data_row.file_paths[self._col_ref.slot_idx]
            return
        elif self.is_error_prop():
            exc = data_row.get_exc(self._col_ref.slot_idx)
            if exc is None:
                data_row[self.slot_idx] = None
            elif self.prop == self.Property.ERRORTYPE:
                data_row[self.slot_idx] = type(exc).__name__
            else:
                data_row[self.slot_idx] = str(exc)
        else:
            raise AssertionError()

    def _as_dict(self) -> dict:
        return {'prop': self.prop.value, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: dict, components: list[Expr]) -> ColumnPropertyRef:
        assert 'prop' in d
        assert isinstance(components[0], ColumnRef)
        return cls(components[0], cls.Property(d['prop']))



from __future__ import annotations

import inspect
import logging
import sys
from textwrap import dedent
from typing import Any, Optional, Sequence, Union

import sqlalchemy as sql

from pixeltable import catalog, exceptions as excs, func, type_system as ts

from .data_row import DataRow
from .expr import Expr
from .literal import Literal
from .row_builder import RowBuilder
from .rowid_ref import RowidRef
from .sql_element_cache import SqlElementCache

_logger = logging.getLogger('pixeltable')


class FunctionCall(Expr):
    fn: func.Function
    is_method_call: bool
    agg_init_args: dict[str, Any]
    resource_pool: Optional[str]

    # These collections hold the component indices corresponding to the args and kwargs
    # that were passed to the FunctionCall. They're 1:1 with the original call pattern.
    arg_idxs: list[int]
    kwarg_idxs: dict[str, int]

    # A "bound" version of the FunctionCall arguments, mapping each specified parameter name
    # to one of three types of bindings:
    # - a component index, if the parameter is a non-variadic parameter
    # - a list of component indices, if the parameter is a variadic positional parameter
    # - a dict mapping keyword names to component indices, if the parameter is a variadic keyword parameter
    bound_idxs: dict[str, Union[int, list[int], dict[str, int]]]

    return_type: ts.ColumnType
    group_by_start_idx: int
    group_by_stop_idx: int
    fn_expr_idx: int
    order_by_start_idx: int
    aggregator: Optional[Any]
    current_partition_vals: Optional[list[Any]]

    _validation_error: Optional[str]

    def __init__(
        self,
        fn: func.Function,
        args: list[Expr],
        kwargs: dict[str, Expr],
        return_type: ts.ColumnType,
        order_by_clause: Optional[list[Any]] = None,
        group_by_clause: Optional[list[Any]] = None,
        is_method_call: bool = False,
        validation_error: Optional[str] = None,
    ):
        assert not fn.is_polymorphic
        assert all(isinstance(arg, Expr) for arg in args)
        assert all(isinstance(arg, Expr) for arg in kwargs.values())

        if order_by_clause is None:
            order_by_clause = []
        if group_by_clause is None:
            group_by_clause = []

        super().__init__(return_type)

        self.fn = fn
        self.return_type = return_type
        self.is_method_call = is_method_call

        # Build the components list from the specified args and kwargs, and note the component_idx of each argument.
        self.components.extend(arg.copy() for arg in args)
        self.arg_idxs = list(range(len(self.components)))
        self.components.extend(arg.copy() for arg in kwargs.values())
        self.kwarg_idxs = {name: i + len(args) for i, name in enumerate(kwargs.keys())}

        # window function state:
        # self.components[self.group_by_start_idx:self.group_by_stop_idx] contains group_by exprs
        self.group_by_start_idx, self.group_by_stop_idx = 0, 0
        if len(group_by_clause) > 0:
            if isinstance(group_by_clause[0], catalog.Table):
                assert len(group_by_clause) == 1
                group_by_exprs = self._create_rowid_refs(group_by_clause[0])
            else:
                assert all(isinstance(expr, Expr) for expr in group_by_clause)
                group_by_exprs = group_by_clause
            # record grouping exprs in self.components, we need to evaluate them to get partition vals
            self.group_by_start_idx = len(self.components)
            self.group_by_stop_idx = len(self.components) + len(group_by_exprs)
            self.components.extend(group_by_exprs)

        if isinstance(self.fn, func.ExprTemplateFunction):
            # we instantiate the template to create an Expr that can be evaluated and record that as a component
            fn_expr = self.fn.instantiate(args, kwargs)
            self.fn_expr_idx = len(self.components)
            self.components.append(fn_expr)
        else:
            self.fn_expr_idx = sys.maxsize

        # we want to make sure that order_by_clause get assigned slot_idxs, even though we won't need to evaluate them
        # (that's done in SQL)
        if len(order_by_clause) > 0 and not isinstance(order_by_clause[0], Expr):
            raise excs.Error(
                f'order_by argument needs to be a Pixeltable expression, but instead is a {type(order_by_clause[0])}'
            )
        self.order_by_start_idx = len(self.components)
        self.components.extend(order_by_clause)

        self._validation_error = validation_error

        if validation_error is not None:
            self.resource_pool = None
            return

        # Now generate bound_idxs for the args and kwargs indices.
        # This is guaranteed to work, because at this point the call has already been validated.
        # These will be used later to dereference specific parameter values.
        bindings = fn.signature.py_signature.bind(*self.arg_idxs, **self.kwarg_idxs)
        self.bound_idxs = bindings.arguments

        # Separately generate bound_args for purposes of determining the resource pool.
        bindings = fn.signature.py_signature.bind(*args, **kwargs)
        bound_args = bindings.arguments
        self.resource_pool = fn.call_resource_pool(bound_args)

        self.agg_init_args = {}
        if self.is_agg_fn_call:
            # We separate out the init args for the aggregator. Unpack Literals in init args.
            assert isinstance(fn, func.AggregateFunction)
            for arg_name, arg in bound_args.items():
                if arg_name in fn.init_param_names[0]:
                    assert isinstance(arg, Literal)  # This was checked during validate_call
                    self.agg_init_args[arg_name] = arg.val

        # execution state for aggregate functions
        self.aggregator = None
        self.current_partition_vals = None

        self.id = self._create_id()

    def _create_rowid_refs(self, tbl: catalog.Table) -> list[Expr]:
        target = tbl._tbl_version_path.tbl_version
        return [RowidRef(target, i) for i in range(target.get().num_rowid_columns())]

    def default_column_name(self) -> Optional[str]:
        return self.fn.name

    def _equals(self, other: FunctionCall) -> bool:
        return (
            self.fn == other.fn
            and self.arg_idxs == other.arg_idxs
            and self.kwarg_idxs == other.kwarg_idxs
            and self.group_by_start_idx == other.group_by_start_idx
            and self.group_by_stop_idx == other.group_by_stop_idx
            and self.order_by_start_idx == other.order_by_start_idx
        )

    def _id_attrs(self) -> list[tuple[str, Any]]:
        return [
            *super()._id_attrs(),
            ('fn', id(self.fn)),  # use the function pointer, not the fqn, which isn't set for lambdas
            ('args', self.arg_idxs),
            ('kwargs', self.kwarg_idxs),
            ('group_by_start_idx', self.group_by_start_idx),
            ('group_by_stop_idx', self.group_by_stop_idx),
            ('fn_expr_idx', self.fn_expr_idx),
            ('order_by_start_idx', self.order_by_start_idx),
        ]

    def __repr__(self) -> str:
        return self.display_str()

    @property
    def validation_error(self) -> Optional[str]:
        return self._validation_error or super().validation_error

    def display_str(self, inline: bool = True) -> str:
        if self.is_method_call:
            return f'{self.components[0]}.{self.fn.name}({self._print_args(1, inline)})'
        else:
            fn_name = self.fn.display_name or 'anonymous_fn'
            return f'{fn_name}({self._print_args()})'

    def _print_args(self, start_idx: int = 0, inline: bool = True) -> str:
        arg_strs = [str(self.components[idx]) for idx in self.arg_idxs[start_idx:]]
        arg_strs.extend([f'{param_name}={self.components[idx]}' for param_name, idx in self.kwarg_idxs.items()])
        if len(self.order_by) > 0:
            assert isinstance(self.fn, func.AggregateFunction)
            if self.fn.requires_order_by:
                arg_strs.insert(0, Expr.print_list(self.order_by))
            else:
                arg_strs.append(f'order_by={Expr.print_list(self.order_by)}')
        if len(self.group_by) > 0:
            arg_strs.append(f'group_by={Expr.print_list(self.group_by)}')
        # TODO: figure out the function name
        separator = ', ' if inline else ',\n    '
        return separator.join(arg_strs)

    def has_group_by(self) -> bool:
        return self.group_by_stop_idx != 0

    @property
    def is_async(self) -> bool:
        return self.fn.is_async

    @property
    def group_by(self) -> list[Expr]:
        return self.components[self.group_by_start_idx : self.group_by_stop_idx]

    @property
    def order_by(self) -> list[Expr]:
        return self.components[self.order_by_start_idx :]

    @property
    def is_window_fn_call(self) -> bool:
        return (
            isinstance(self.fn, func.AggregateFunction)
            and self.fn.allows_window
            and (
                not self.fn.allows_std_agg
                or self.has_group_by()
                or (len(self.order_by) > 0 and not self.fn.requires_order_by)
            )
        )

    def get_window_sort_exprs(self) -> tuple[list[Expr], list[Expr]]:
        return self.group_by, self.order_by

    def get_window_ordering(self) -> list[tuple[Expr, bool]]:
        # ordering is implicitly ascending
        return [(e, None) for e in self.group_by] + [(e, True) for e in self.order_by]

    @property
    def is_agg_fn_call(self) -> bool:
        return isinstance(self.fn, func.AggregateFunction)

    def get_agg_order_by(self) -> list[Expr]:
        assert self.is_agg_fn_call
        return self.order_by

    def sql_expr(self, sql_elements: SqlElementCache) -> Optional[sql.ColumnElement]:
        assert self.is_valid

        # we currently can't translate aggregate functions with grouping and/or ordering to SQL
        if self.has_group_by() or len(self.order_by) > 0:
            return None

        # try to construct args and kwargs to call self.fn._to_sql()
        args: list[sql.ColumnElement] = []
        for component_idx in self.arg_idxs:
            arg_element = sql_elements.get(self.components[component_idx])
            if arg_element is None:
                return None
            args.append(arg_element)

        kwargs: dict[str, sql.ColumnElement] = {}
        for param_name, component_idx in self.kwarg_idxs.items():
            arg_element = sql_elements.get(self.components[component_idx])
            if arg_element is None:
                return None
            kwargs[param_name] = arg_element

        return self.fn._to_sql(*args, **kwargs)

    def reset_agg(self) -> None:
        """
        Init agg state
        """
        assert self.is_agg_fn_call
        assert isinstance(self.fn, func.AggregateFunction)
        self.aggregator = self.fn.agg_class(**self.agg_init_args)

    @property
    def bound_args(self) -> dict[str, Expr]:
        """
        Reconstructs bound arguments from the components of this FunctionCall.
        """
        bound_args: dict[str, Expr] = {}
        for name, idx in self.bound_idxs.items():
            if isinstance(idx, int):
                bound_args[name] = self.components[idx]
            elif isinstance(idx, Sequence):
                bound_args[name] = Expr.from_object([self.components[i] for i in idx])
            elif isinstance(idx, dict):
                bound_args[name] = Expr.from_object({k: self.components[i] for k, i in idx.items()})
            else:
                raise AssertionError(f'{name}: {idx} (of type `{type(idx)}`)')
        return bound_args

    def substitute(self, spec: dict[Expr, Expr]) -> Expr:
        """
        Substitution of FunctionCall arguments could cause the return value to become more specific, in the case
        where a variable is replaced with a specific value.
        """
        res = super().substitute(spec)
        assert res is self
        self.return_type = self.fn.call_return_type(self.bound_args)
        self.col_type = self.return_type
        return self

    def update(self, data_row: DataRow) -> None:
        """
        Update agg state
        """
        assert self.is_agg_fn_call
        args, kwargs = self.make_args(data_row)
        self.aggregator.update(*args, **kwargs)

    def make_args(self, data_row: DataRow) -> Optional[tuple[list[Any], dict[str, Any]]]:
        """Return args and kwargs, constructed for data_row; returns None if any non-nullable arg is None."""
        args: list[Any] = []
        parameters_by_pos = self.fn.signature.parameters_by_pos
        for idx in self.arg_idxs:
            val = data_row[self.components[idx].slot_idx]
            if (
                val is None
                and parameters_by_pos[idx].kind
                in (inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD)
                and not parameters_by_pos[idx].col_type.nullable
            ):
                return None
            args.append(val)

        kwargs: dict[str, Any] = {}
        parameters = self.fn.signature.parameters
        for param_name, idx in self.kwarg_idxs.items():
            val = data_row[self.components[idx].slot_idx]
            if (
                val is None
                and parameters[param_name].kind
                in (inspect.Parameter.KEYWORD_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD)
                and not parameters[param_name].col_type.nullable
            ):
                return None
            kwargs[param_name] = val

        return args, kwargs

    def get_param_values(self, param_names: Sequence[str], data_rows: list[DataRow]) -> list[dict[str, Any]]:
        """
        Returns a list of dicts mapping each param name to its value when this FunctionCall is evaluated against
        data_rows
        """
        assert self.is_valid
        assert all(name in self.fn.signature.parameters for name in param_names), f'{param_names}, {self.fn.signature}'
        result: list[dict[str, Any]] = []
        for row in data_rows:
            d: dict[str, Any] = {}
            for param_name in param_names:
                val = self.bound_idxs.get(param_name)
                if isinstance(val, int):
                    d[param_name] = row[self.components[val].slot_idx]
                elif isinstance(val, list):
                    # var_positional
                    d[param_name] = [row[self.components[idx].slot_idx] for idx in val]
                elif isinstance(val, dict):
                    # var_keyword
                    d[param_name] = {k: row[self.components[idx].slot_idx] for k, idx in val.items()}
                else:
                    assert val is None
                    default = self.fn.signature.parameters[param_name].default
                    assert default is not None
                    d[param_name] = default.val
            result.append(d)
        return result

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        assert self.is_valid

        if isinstance(self.fn, func.ExprTemplateFunction):
            # we need to evaluate the template
            # TODO: can we get rid of this extra copy?
            fn_expr = self.components[self.fn_expr_idx]
            data_row[self.slot_idx] = data_row[fn_expr.slot_idx]
            return
        elif self.is_agg_fn_call and not self.is_window_fn_call:
            if self.aggregator is None:
                pass
            data_row[self.slot_idx] = self.aggregator.value()
            return

        args_kwargs = self.make_args(data_row)
        if args_kwargs is None:
            # we can't evaluate this function
            data_row[self.slot_idx] = None
            return
        args, kwargs = args_kwargs

        if self.is_window_fn_call:
            assert isinstance(self.fn, func.AggregateFunction)
            agg_cls = self.fn.agg_class
            if self.has_group_by():
                if self.current_partition_vals is None:
                    self.current_partition_vals = [None] * len(self.group_by)
                partition_vals = [data_row[e.slot_idx] for e in self.group_by]
                if partition_vals != self.current_partition_vals:
                    # new partition
                    self.aggregator = agg_cls(**self.agg_init_args)
                    self.current_partition_vals = partition_vals
            elif self.aggregator is None:
                self.aggregator = agg_cls(**self.agg_init_args)
            self.aggregator.update(*args)
            data_row[self.slot_idx] = self.aggregator.value()
        else:
            data_row[self.slot_idx] = self.fn.exec(args, kwargs)

    def _as_dict(self) -> dict:
        return {
            'fn': self.fn.as_dict(),
            'return_type': self.return_type.as_dict(),
            'arg_idxs': self.arg_idxs,
            'kwarg_idxs': self.kwarg_idxs,
            'group_by_start_idx': self.group_by_start_idx,
            'group_by_stop_idx': self.group_by_stop_idx,
            'order_by_start_idx': self.order_by_start_idx,
            'is_method_call': self.is_method_call,
            **super()._as_dict(),
        }

    @classmethod
    def _from_dict(cls, d: dict, components: list[Expr]) -> FunctionCall:
        fn = func.Function.from_dict(d['fn'])
        return_type = ts.ColumnType.from_dict(d['return_type']) if 'return_type' in d else None
        arg_idxs: list[int] = d['arg_idxs']
        kwarg_idxs: dict[str, int] = d['kwarg_idxs']
        group_by_start_idx: int = d['group_by_start_idx']
        group_by_stop_idx: int = d['group_by_stop_idx']
        order_by_start_idx: int = d['order_by_start_idx']
        is_method_call: bool = d['is_method_call']

        args = [components[idx] for idx in arg_idxs]
        kwargs = {name: components[idx] for name, idx in kwarg_idxs.items()}
        group_by_exprs = components[group_by_start_idx:group_by_stop_idx]
        order_by_exprs = components[order_by_start_idx:]

        validation_error: Optional[str] = None

        if isinstance(fn, func.InvalidFunction):
            validation_error = (
                dedent(
                    f"""
                    The UDF '{fn.self_path}' cannot be located, because
                    {{errormsg}}
                    """
                )
                .strip()
                .format(errormsg=fn.errormsg)
            )
            return cls(fn, args, kwargs, return_type, is_method_call=is_method_call, validation_error=validation_error)

        # Now re-bind args and kwargs using the version of `fn` that is currently represented in code. This ensures
        # that we get a valid binding even if the signatures of `fn` have changed since the FunctionCall was
        # serialized.

        resolved_fn: func.Function = fn

        try:
            # Bind args and kwargs to the function signature in the current codebase.
            resolved_fn, bound_args = fn._bind_to_matching_signature(args, kwargs)
        except (TypeError, excs.Error):
            signature_note_str = 'any of its signatures' if fn.is_polymorphic else 'its signature'
            args_str = [str(arg.col_type) for arg in args]
            args_str.extend(f'{name}: {arg.col_type}' for name, arg in kwargs.items())
            call_signature_str = f'({", ".join(args_str)}) -> {return_type}'
            fn_signature_str = f'{len(fn.signatures)} signatures' if fn.is_polymorphic else str(fn.signature)
            validation_error = dedent(
                f"""
                The signature stored in the database for a UDF call to {fn.self_path!r} no longer
                matches {signature_note_str} as currently defined in the code. This probably means that the
                code for {fn.self_path!r} has changed in a backward-incompatible way.
                Signature of UDF call in the database: {call_signature_str}
                Signature of UDF as currently defined in code: {fn_signature_str}
                """
            ).strip()
        else:
            # Evaluate the call_return_type as defined in the current codebase.
            call_return_type = resolved_fn.call_return_type(bound_args)
            if return_type is None:
                # Schema versions prior to 25 did not store the return_type in metadata, and there is no obvious way to
                # infer it during DB migration, so we might encounter a stored return_type of None. In that case, we use
                # the call_return_type that we just inferred (which matches the deserialization behavior prior to
                # version 25).
                return_type = call_return_type
            elif not return_type.is_supertype_of(call_return_type, ignore_nullable=True):
                # There is a return_type stored in metadata (schema version >= 25),
                # and the stored return_type of the UDF call doesn't match the column type of the FunctionCall.
                validation_error = dedent(
                    f"""
                    The return type stored in the database for a UDF call to {fn.self_path!r} no longer
                    matches its return type as currently defined in the code. This probably means that the
                    code for {fn.self_path!r} has changed in a backward-incompatible way.
                    Return type of UDF call in the database: {return_type}
                    Return type of UDF as currently defined in code: {call_return_type}
                    """
                ).strip()

        fn_call = cls(
            resolved_fn,
            args,
            kwargs,
            return_type,
            group_by_clause=group_by_exprs,
            order_by_clause=order_by_exprs,
            is_method_call=is_method_call,
            validation_error=validation_error,
        )

        return fn_call



from __future__ import annotations

import logging
from typing import Optional

import numpy as np

from pixeltable import exprs

_logger = logging.getLogger('pixeltable')


class RowBuffer:
    """Fixed-length circular buffer of DataRows; knows how to maintain input order"""

    size: int
    row_pos_map: Optional[dict[int, int]]  # id(row) -> position of row in output; None if not maintaining order
    num_rows: int  # number of rows in the buffer
    num_ready: int  # number of consecutive non-None rows at head
    buffer: np.ndarray  # of object
    head_idx: int  # index of beginning of the buffer
    head_pos: int  # row position of the beginning of the buffer

    def __init__(self, size: int):
        self.size = size
        self.row_pos_map = None
        self.num_rows = 0
        self.num_ready = 0
        self.buffer = np.full(size, None, dtype=object)
        self.head_pos = 0
        self.head_idx = 0

    def set_row_pos_map(self, row_pos_map: dict[int, int]) -> None:
        self.row_pos_map = row_pos_map

    def add_row(self, row: exprs.DataRow) -> None:
        offset: int  # of new row from head
        if self.row_pos_map is not None:
            pos = self.row_pos_map.get(id(row))
            assert pos is not None and (pos - self.head_pos < self.size), f'{pos} {self.head_pos} {self.size}'
            offset = pos - self.head_pos
        else:
            offset = self.num_rows
        idx = (self.head_idx + offset) % self.size
        assert self.buffer[idx] is None

        self.buffer[idx] = row
        self.num_rows += 1
        if self.row_pos_map is not None:
            if offset == self.num_ready:
                # we have new ready rows; find out how many
                while offset < self.size and self.buffer[(self.head_idx + offset) % self.size] is not None:
                    offset += 1
                self.num_ready = offset
        else:
            self.num_ready += 1

    def get_rows(self, n: int) -> list[exprs.DataRow]:
        """Get up to n ready rows from head"""
        n = min(n, self.num_ready)
        if n == 0:
            return []
        rows: list[exprs.DataRow]
        if self.head_idx + n <= self.size:
            rows = self.buffer[self.head_idx : self.head_idx + n].tolist()
            self.buffer[self.head_idx : self.head_idx + n] = None
        else:
            rows = np.concatenate([self.buffer[self.head_idx :], self.buffer[: self.head_idx + n - self.size]]).tolist()
            self.buffer[self.head_idx :] = None
            self.buffer[: self.head_idx + n - self.size] = None
        self.head_pos += n
        self.head_idx = (self.head_idx + n) % self.size
        self.num_rows -= n
        self.num_ready -= n
        return rows



from __future__ import annotations

import asyncio
import datetime
import inspect
import logging
import sys
import time
from typing import Awaitable, Collection, Optional

from pixeltable import env, func
from pixeltable.config import Config

from .globals import Dispatcher, ExecCtx, FnCallArgs, Scheduler

_logger = logging.getLogger('pixeltable')


class RateLimitsScheduler(Scheduler):
    """
    Scheduler for FunctionCalls with a RateLimitsInfo pool, which provides information about actual resource usage.

    Scheduling strategy:
    - try to stay below resource limits by utilizing reported RateLimitInfo.remaining
    - also take into account the estimated resource usage for in-flight requests
      (obtained via RateLimitsInfo.get_request_resources())
    - issue synchronous requests when we don't have a RateLimitsInfo yet or when we depleted a resource and need to
      wait for a reset

    TODO:
    - limit the number of in-flight requests based on the open file limit
    """

    get_request_resources_param_names: list[str]  # names of parameters of RateLimitsInfo.get_request_resources()

    # scheduling-related state
    pool_info: Optional[env.RateLimitsInfo]
    est_usage: dict[str, int]  # value per resource; accumulated estimates since the last util. report

    num_in_flight: int  # unfinished tasks
    request_completed: asyncio.Event

    total_requests: int
    total_retried: int

    TIME_FORMAT = '%H:%M.%S %f'
    MAX_RETRIES = 10

    def __init__(self, resource_pool: str, dispatcher: Dispatcher):
        super().__init__(resource_pool, dispatcher)
        loop_task = asyncio.create_task(self._main_loop())
        self.dispatcher.register_task(loop_task)
        self.pool_info = None  # initialized in _main_loop by the first request
        self.est_usage = {}
        self.num_in_flight = 0
        self.request_completed = asyncio.Event()
        self.total_requests = 0
        self.total_retried = 0
        self.get_request_resources_param_names = []

    @classmethod
    def matches(cls, resource_pool: str) -> bool:
        return resource_pool.startswith('rate-limits:')

    def _set_pool_info(self) -> None:
        """Initialize pool_info with the RateLimitsInfo for the resource pool, if available"""
        if self.pool_info is not None:
            return
        self.pool_info = env.Env.get().get_resource_pool_info(self.resource_pool, None)
        if self.pool_info is None:
            return
        assert isinstance(self.pool_info, env.RateLimitsInfo)
        assert hasattr(self.pool_info, 'get_request_resources')
        sig = inspect.signature(self.pool_info.get_request_resources)
        self.get_request_resources_param_names = [p.name for p in sig.parameters.values()]
        self.est_usage = dict.fromkeys(self._resources, 0)

    async def _main_loop(self) -> None:
        item: Optional[RateLimitsScheduler.QueueItem] = None
        while True:
            if item is None:
                item = await self.queue.get()
                if item.num_retries > 0:
                    self.total_retried += 1

            now = datetime.datetime.now(tz=datetime.timezone.utc)
            if self.pool_info is None or not self.pool_info.is_initialized():
                # wait for a single request to get rate limits
                _logger.debug(f'initializing rate limits for {self.resource_pool}')
                await self._exec(item.request, item.exec_ctx, item.num_retries, is_task=False)
                _logger.debug(f'initialized rate limits for {self.resource_pool}')
                item = None
                # if this was the first request, it created the pool_info
                if self.pool_info is None:
                    self._set_pool_info()
                continue

            # check rate limits
            _logger.debug(f'checking rate limits for {self.resource_pool}')
            request_resources = self._get_request_resources(item.request)
            limits_info = self._check_resource_limits(request_resources)
            aws: list[Awaitable[None]] = []
            completed_aw: Optional[asyncio.Task] = None
            wait_for_reset: Optional[asyncio.Task] = None
            if limits_info is not None:
                # limits_info's resource is depleted, wait for capacity to free up

                if self.num_in_flight > 0:
                    # a completed request can free up capacity
                    self.request_completed.clear()
                    completed_aw = asyncio.create_task(self.request_completed.wait())
                    aws.append(completed_aw)
                    _logger.debug(f'waiting for completed request for {self.resource_pool}')

                reset_at = limits_info.reset_at
                if reset_at > now:
                    # we're waiting for the rate limit to reset
                    wait_for_reset = asyncio.create_task(asyncio.sleep((reset_at - now).total_seconds()))
                    aws.append(wait_for_reset)
                    _logger.debug(f'waiting for rate limit reset for {self.resource_pool}')

            if len(aws) > 0:
                # we have something to wait for
                done, pending = await asyncio.wait(aws, return_when=asyncio.FIRST_COMPLETED)
                for task in pending:
                    task.cancel()
                if completed_aw in done:
                    _logger.debug(f'wait(): completed request for {self.resource_pool}')
                if wait_for_reset in done:
                    _logger.debug(f'wait(): rate limit reset for {self.resource_pool}')
                    # force waiting for another rate limit report before making any scheduling decisions
                    self.pool_info.reset()
                # re-evaluate current capacity for current item
                continue

            # we have a new in-flight request
            for resource, val in request_resources.items():
                self.est_usage[resource] += val
            _logger.debug(f'creating task for {self.resource_pool}')
            self.num_in_flight += 1
            task = asyncio.create_task(self._exec(item.request, item.exec_ctx, item.num_retries, is_task=True))
            self.dispatcher.register_task(task)
            item = None

    @property
    def _resources(self) -> Collection[str]:
        return self.pool_info.resource_limits.keys() if self.pool_info is not None else []

    def _get_request_resources(self, request: FnCallArgs) -> dict[str, int]:
        kwargs_batch = request.fn_call.get_param_values(self.get_request_resources_param_names, request.rows)
        if not request.is_batched:
            return self.pool_info.get_request_resources(**kwargs_batch[0])
        else:
            batch_kwargs = {k: [d[k] for d in kwargs_batch] for k in kwargs_batch[0]}
            constant_kwargs, batch_kwargs = request.pxt_fn.create_batch_kwargs(batch_kwargs)
            return self.pool_info.get_request_resources(**constant_kwargs, **batch_kwargs)

    def _check_resource_limits(self, request_resources: dict[str, int]) -> Optional[env.RateLimitInfo]:
        """Returns the most depleted resource, relative to its limit, or None if all resources are within limits"""
        candidates: list[tuple[env.RateLimitInfo, float]] = []  # (info, relative usage)
        for resource, usage in request_resources.items():
            # 0.05: leave some headroom, we don't have perfect information
            info = self.pool_info.resource_limits[resource]
            est_remaining = info.remaining - self.est_usage[resource] - usage
            if est_remaining < 0.05 * info.limit:
                candidates.append((info, est_remaining / info.limit))
        if len(candidates) == 0:
            return None
        return min(candidates, key=lambda x: x[1])[0]

    async def _exec(self, request: FnCallArgs, exec_ctx: ExecCtx, num_retries: int, is_task: bool) -> None:
        assert all(not row.has_val[request.fn_call.slot_idx] for row in request.rows)
        assert all(not row.has_exc(request.fn_call.slot_idx) for row in request.rows)

        try:
            start_ts = datetime.datetime.now(tz=datetime.timezone.utc)
            pxt_fn = request.fn_call.fn
            assert isinstance(pxt_fn, func.CallableFunction)
            _logger.debug(
                f'scheduler {self.resource_pool}: '
                f'start evaluating slot {request.fn_call.slot_idx}, batch_size={len(request.rows)}'
            )
            self.total_requests += 1
            if request.is_batched:
                batch_result = await pxt_fn.aexec_batch(*request.batch_args, **request.batch_kwargs)
                assert len(batch_result) == len(request.rows)
                for row, result in zip(request.rows, batch_result):
                    row[request.fn_call.slot_idx] = result
            else:
                result = await pxt_fn.aexec(*request.args, **request.kwargs)
                request.row[request.fn_call.slot_idx] = result
            end_ts = datetime.datetime.now(tz=datetime.timezone.utc)
            _logger.debug(
                f'scheduler {self.resource_pool}: evaluated slot {request.fn_call.slot_idx} '
                f'in {end_ts - start_ts}, batch_size={len(request.rows)}'
            )

            # purge accumulated usage estimate, now that we have a new report
            self.est_usage = dict.fromkeys(self._resources, 0)

            self.dispatcher.dispatch(request.rows, exec_ctx)
        except Exception as exc:
            _logger.debug(f'scheduler {self.resource_pool}: exception in slot {request.fn_call.slot_idx}: {exc}')
            if self.pool_info is None:
                # our pool info should be available at this point
                self._set_pool_info()
            assert self.pool_info is not None
            if num_retries < self.MAX_RETRIES:
                retry_delay = self.pool_info.get_retry_delay(exc)
                if retry_delay is not None:
                    self.total_retried += 1
                    _logger.debug(f'scheduler {self.resource_pool}: retrying in {retry_delay} seconds')
                    await asyncio.sleep(retry_delay)
                    self.queue.put_nowait(self.QueueItem(request, num_retries + 1, exec_ctx))
                    return
            # TODO: update resource limits reported in exc.response.headers, if present

            # record the exception
            _, _, exc_tb = sys.exc_info()
            for row in request.rows:
                row.set_exc(request.fn_call.slot_idx, exc)
            self.dispatcher.dispatch_exc(request.rows, request.fn_call.slot_idx, exc_tb, exec_ctx)
        finally:
            _logger.debug(f'Scheduler stats: #requests={self.total_requests}, #retried={self.total_retried}')
            if is_task:
                self.num_in_flight -= 1
                self.request_completed.set()


class RequestRateScheduler(Scheduler):
    """
    Scheduler for FunctionCalls with a fixed request rate limit and no runtime resource usage reports.

    Rate limits are supplied in the config, in one of two ways:
    - resource_pool='request-rate:<endpoint>':
      * a single rate limit for all calls against that endpoint
      * in the config: section '<endpoint>', key 'rate_limit'
    - resource_pool='request-rate:<endpoint>:<model>':
        * a single rate limit for all calls against that model
        * in the config: section '<endpoint>.rate_limits', key '<model>'
    - if no rate limit is found in the config, uses a default of 600 RPM

    TODO:
    - adaptive rate limiting based on 429 errors
    """

    secs_per_request: float  # inverted rate limit
    num_in_flight: int
    total_requests: int
    total_retried: int

    TIME_FORMAT = '%H:%M.%S %f'
    MAX_RETRIES = 10
    DEFAULT_RATE_LIMIT = 600  # requests per minute

    def __init__(self, resource_pool: str, dispatcher: Dispatcher):
        super().__init__(resource_pool, dispatcher)
        loop_task = asyncio.create_task(self._main_loop())
        self.dispatcher.register_task(loop_task)
        self.num_in_flight = 0
        self.total_requests = 0
        self.total_retried = 0

        # try to get the rate limit from the config
        elems = resource_pool.split(':')
        section: str
        key: str
        if len(elems) == 2:
            # resource_pool: request-rate:endpoint
            _, endpoint = elems
            section = endpoint
            key = 'rate_limit'
        else:
            # resource_pool: request-rate:endpoint:model
            assert len(elems) == 3
            _, endpoint, model = elems
            section = f'{endpoint}.rate_limits'
            key = model
        requests_per_min = Config.get().get_int_value(key, section=section)
        requests_per_min = requests_per_min or self.DEFAULT_RATE_LIMIT
        self.secs_per_request = 1 / (requests_per_min / 60)

    @classmethod
    def matches(cls, resource_pool: str) -> bool:
        return resource_pool.startswith('request-rate:')

    async def _main_loop(self) -> None:
        last_request_ts = 0.0
        while True:
            item = await self.queue.get()
            if item.num_retries > 0:
                self.total_retried += 1
            now = time.monotonic()
            if now - last_request_ts < self.secs_per_request:
                wait_duration = self.secs_per_request - (now - last_request_ts)
                _logger.debug(f'waiting for {wait_duration} for {self.resource_pool}')
                await asyncio.sleep(wait_duration)

            last_request_ts = time.monotonic()
            if item.num_retries > 0:
                # the last request encountered some problem: retry it synchronously, to wait for the problem to pass
                _logger.debug(f'retrying request for {self.resource_pool}: #retries={item.num_retries}')
                await self._exec(item.request, item.exec_ctx, item.num_retries, is_task=False)
                _logger.debug(f'retried request for {self.resource_pool}: #retries={item.num_retries}')
            else:
                _logger.debug(f'creating task for {self.resource_pool}')
                self.num_in_flight += 1
                task = asyncio.create_task(self._exec(item.request, item.exec_ctx, item.num_retries, is_task=True))
                self.dispatcher.register_task(task)

    async def _exec(self, request: FnCallArgs, exec_ctx: ExecCtx, num_retries: int, is_task: bool) -> None:
        assert all(not row.has_val[request.fn_call.slot_idx] for row in request.rows)
        assert all(not row.has_exc(request.fn_call.slot_idx) for row in request.rows)

        try:
            start_ts = datetime.datetime.now(tz=datetime.timezone.utc)
            pxt_fn = request.fn_call.fn
            assert isinstance(pxt_fn, func.CallableFunction)
            _logger.debug(
                f'scheduler {self.resource_pool}: '
                f'start evaluating slot {request.fn_call.slot_idx}, batch_size={len(request.rows)}'
            )
            self.total_requests += 1
            if request.is_batched:
                batch_result = await pxt_fn.aexec_batch(*request.batch_args, **request.batch_kwargs)
                assert len(batch_result) == len(request.rows)
                for row, result in zip(request.rows, batch_result):
                    row[request.fn_call.slot_idx] = result
            else:
                result = await pxt_fn.aexec(*request.args, **request.kwargs)
                request.row[request.fn_call.slot_idx] = result
            end_ts = datetime.datetime.now(tz=datetime.timezone.utc)
            _logger.debug(
                f'scheduler {self.resource_pool}: evaluated slot {request.fn_call.slot_idx} '
                f'in {end_ts - start_ts}, batch_size={len(request.rows)}'
            )
            self.dispatcher.dispatch(request.rows, exec_ctx)

        except Exception as exc:
            # TODO: which exception can be retried?
            _logger.debug(f'exception for {self.resource_pool}: {exc}')
            status = getattr(exc, 'status', None)
            _logger.debug(f'type={type(exc)} has_status={hasattr(exc, "status")} status={status}')
            if num_retries < self.MAX_RETRIES:
                self.queue.put_nowait(self.QueueItem(request, num_retries + 1, exec_ctx))
                return

            # record the exception
            _, _, exc_tb = sys.exc_info()
            for row in request.rows:
                row.set_exc(request.fn_call.slot_idx, exc)
            self.dispatcher.dispatch_exc(request.rows, request.fn_call.slot_idx, exc_tb, exec_ctx)
        finally:
            _logger.debug(
                f'Scheduler stats: #in-flight={self.num_in_flight} #requests={self.total_requests}, '
                f'#retried={self.total_retried}'
            )
            if is_task:
                self.num_in_flight -= 1


# all concrete Scheduler subclasses that implement matches()
SCHEDULERS = [RateLimitsScheduler, RequestRateScheduler]



from __future__ import annotations

import logging
from typing import Optional
from uuid import UUID

from pixeltable.metadata import schema

from .column import Column
from .table_version_handle import TableVersionHandle

_logger = logging.getLogger('pixeltable')


class TableVersionPath:
    """
    A TableVersionPath represents the sequence of TableVersions from a base table to a particular view:
    - for a base table: only includes that TableVersion
    - for a view: includes the TableVersion for the view and all its bases
    - multiple snapshots can reference the same TableVersion, but with different bases, which means that the
      graph of TableVersions is a DAG, not a tree (which is why we cannot embed the DAG into TableVersion directly)

    TableVersionPath contains all metadata needed to execute queries and updates against a particular version of a
    table/view.
    """

    tbl_version: TableVersionHandle
    base: Optional[TableVersionPath]

    def __init__(self, tbl_version: TableVersionHandle, base: Optional[TableVersionPath] = None):
        assert tbl_version is not None
        self.tbl_version = tbl_version
        self.base = base

    @classmethod
    def from_md(cls, path: schema.TableVersionPath) -> TableVersionPath:
        assert len(path) > 0
        result: Optional[TableVersionPath] = None
        for tbl_id_str, effective_version in path[::-1]:
            tbl_id = UUID(tbl_id_str)
            result = TableVersionPath(TableVersionHandle(tbl_id, effective_version), base=result)
        return result

    def as_md(self) -> schema.TableVersionPath:
        result = [(self.tbl_version.id.hex, self.tbl_version.effective_version)]
        if self.base is not None:
            result.extend(self.base.as_md())
        return result

    def tbl_id(self) -> UUID:
        """Return the id of the table/view that this path represents"""
        return self.tbl_version.id

    def version(self) -> int:
        """Return the version of the table/view that this path represents"""
        return self.tbl_version.get().version

    def tbl_name(self) -> str:
        """Return the name of the table/view that this path represents"""
        return self.tbl_version.get().name

    def path_len(self) -> int:
        """Return the length of the path"""
        return 1 if self.base is None else 1 + self.base.path_len()

    def is_snapshot(self) -> bool:
        """Return True if this is a path of snapshot versions"""
        if not self.tbl_version.get().is_snapshot:
            return False
        return self.base.is_snapshot() if self.base is not None else True

    def is_view(self) -> bool:
        return self.tbl_version.get().is_view

    def is_component_view(self) -> bool:
        return self.tbl_version.get().is_component_view

    def is_insertable(self) -> bool:
        return self.tbl_version.get().is_insertable()

    def get_tbl_versions(self) -> list[TableVersionHandle]:
        """Return all tbl versions"""
        if self.base is None:
            return [self.tbl_version]
        return [self.tbl_version, *self.base.get_tbl_versions()]

    def get_bases(self) -> list[TableVersionHandle]:
        """Return all tbl versions"""
        if self.base is None:
            return []
        return self.base.get_tbl_versions()

    def find_tbl_version(self, id: UUID) -> Optional[TableVersionHandle]:
        """Return the matching TableVersion in the chain of TableVersions, starting with this one"""
        if self.tbl_version.id == id:
            return self.tbl_version
        if self.base is None:
            return None
        return self.base.find_tbl_version(id)

    def columns(self) -> list[Column]:
        """Return all user columns visible in this tbl version path, including columns from bases"""
        result = list(self.tbl_version.get().cols_by_name.values())
        if self.base is not None and self.tbl_version.get().include_base_columns:
            base_cols = self.base.columns()
            # we only include base columns that don't conflict with one of our column names
            result.extend(c for c in base_cols if c.name not in self.tbl_version.get().cols_by_name)
        return result

    def cols_by_name(self) -> dict[str, Column]:
        """Return a dict of all user columns visible in this tbl version path, including columns from bases"""
        cols = self.columns()
        return {col.name: col for col in cols}

    def cols_by_id(self) -> dict[int, Column]:
        """Return a dict of all user columns visible in this tbl version path, including columns from bases"""
        cols = self.columns()
        return {col.id: col for col in cols}

    def get_column(self, name: str, include_bases: Optional[bool] = None) -> Optional[Column]:
        """Return the column with the given name, or None if not found"""
        col = self.tbl_version.get().cols_by_name.get(name)
        if col is not None:
            return col
        elif self.base is not None and (include_bases or self.tbl_version.get().include_base_columns):
            return self.base.get_column(name)
        else:
            return None

    def get_column_by_id(self, tbl_id: UUID, col_id: int) -> Optional[Column]:
        """Return the column for the given tbl/col id"""
        if self.tbl_version.id == tbl_id:
            assert col_id in self.tbl_version.get().cols_by_id
            return self.tbl_version.get().cols_by_id[col_id]
        elif self.base is not None:
            return self.base.get_column_by_id(tbl_id, col_id)
        else:
            return None

    def has_column(self, col: Column, include_bases: bool = True) -> bool:
        """Return True if this table has the given column."""
        assert col.tbl is not None
        if (
            col.tbl.id == self.tbl_version.id
            and col.tbl.effective_version == self.tbl_version.effective_version
            and col.id in self.tbl_version.get().cols_by_id
        ):
            # the column is visible in this table version
            return True
        elif self.base is not None and include_bases:
            return self.base.has_column(col)
        else:
            return False

    def as_dict(self) -> dict:
        return {
            'tbl_version': self.tbl_version.as_dict(),
            'base': self.base.as_dict() if self.base is not None else None,
        }

    @classmethod
    def from_dict(cls, d: dict) -> TableVersionPath:
        tbl_version = TableVersionHandle.from_dict(d['tbl_version'])
        base = TableVersionPath.from_dict(d['base']) if d['base'] is not None else None
        return cls(tbl_version, base)



from __future__ import annotations

import sys
import time
from dataclasses import dataclass
from typing import Any, Iterable, Optional, Sequence
from uuid import UUID

import numpy as np

from pixeltable import catalog, exceptions as excs, utils
from pixeltable.env import Env
from pixeltable.utils.media_store import MediaStore

from .data_row import DataRow
from .expr import Expr, ExprScope
from .expr_set import ExprSet


class ExecProfile:
    def __init__(self, row_builder: RowBuilder):
        self.eval_time = [0.0] * row_builder.num_materialized
        self.eval_count = [0] * row_builder.num_materialized
        self.row_builder = row_builder

    def print(self, num_rows: int) -> None:
        for i in range(self.row_builder.num_materialized):
            if self.eval_count[i] == 0:
                continue
            per_call_time = self.eval_time[i] / self.eval_count[i]
            calls_per_row = self.eval_count[i] / num_rows
            multiple_str = f'({calls_per_row}x)' if calls_per_row > 1 else ''
            Env.get().console_logger.info(
                f'{self.row_builder.unique_exprs[i]}: {utils.print_perf_counter_delta(per_call_time)} {multiple_str}'
            )


@dataclass
class ColumnSlotIdx:
    """Info for how to locate materialized column in DataRow
    TODO: can this be integrated into RowBuilder directly?
    """

    col: catalog.Column
    slot_idx: int


class RowBuilder:
    """Create and populate DataRows and table rows from exprs and computed columns

    For ColumnRefs to unstored iterator columns:
    - in order for them to be executable, we also record the iterator args and pass them to the ColumnRef
    """

    unique_exprs: ExprSet
    next_slot_idx: int
    input_expr_slot_idxs: set[int]

    # output exprs: all exprs the caller wants to materialize
    # - explicitly requested output_exprs
    # - values for computed columns
    output_exprs: ExprSet

    input_exprs: ExprSet

    table_columns: list[ColumnSlotIdx]
    default_eval_ctx: EvalCtx
    unstored_iter_args: dict[UUID, Expr]

    # transitive dependents for the purpose of exception propagation: an exception for slot i is propagated to
    # _exc_dependents[i]
    # (list of set of slot_idxs, indexed by slot_idx)
    _exc_dependents: list[set[int]]

    # dependents[i] = direct dependents of expr with slot idx i; dependents[i, j] == True: expr j depends on expr i
    dependents: np.ndarray  # of bool
    transitive_dependents: np.ndarray  # of bool
    # dependencies[i] = direct dependencies of expr with slot idx i; transpose of dependents
    dependencies: np.ndarray  # of bool
    # num_dependencies[i] = number of direct dependencies of expr with slot idx i
    num_dependencies: np.ndarray  # of int

    # records the output_expr that a subexpr belongs to
    # (a subexpr can be shared across multiple output exprs)
    output_expr_ids: list[set[int]]

    @dataclass
    class EvalCtx:
        """Context for evaluating a set of target exprs"""

        slot_idxs: list[int]  # slot idxs of exprs needed to evaluate target exprs; does not contain duplicates
        exprs: list[Expr]  # exprs corresponding to slot_idxs
        target_slot_idxs: list[int]  # slot idxs of target exprs; might contain duplicates
        target_exprs: list[Expr]  # exprs corresponding to target_slot_idxs

    def __init__(self, output_exprs: Sequence[Expr], columns: Sequence[catalog.Column], input_exprs: Iterable[Expr]):
        """
        Args:
            output_exprs: list of Exprs to be evaluated
            columns: list of columns to be materialized
            input_exprs: list of Exprs that are excluded from evaluation (because they're already materialized)
        TODO: enforce that output_exprs doesn't overlap with input_exprs?
        """
        self.unique_exprs: ExprSet[Expr] = ExprSet()  # dependencies precede their dependents
        self.next_slot_idx = 0

        # record input and output exprs; make copies to avoid reusing execution state
        unique_input_exprs = [self._record_unique_expr(e.copy(), recursive=False) for e in input_exprs]
        self.input_expr_slot_idxs = {e.slot_idx for e in unique_input_exprs}

        resolve_cols = set(columns)
        self.output_exprs = ExprSet(
            [
                self._record_unique_expr(e.copy().resolve_computed_cols(resolve_cols=resolve_cols), recursive=True)
                for e in output_exprs
            ]
        )

        # if init(columns):
        # - we are creating table rows and need to record columns for create_table_row()
        # - output_exprs materialize those columns
        # - input_exprs are ColumnRefs of the non-computed columns (ie, what needs to be provided as input)
        # - media validation:
        #   * for write-validated columns, we need to create validating ColumnRefs
        #   * further references to that column (eg, computed cols) need to resolve to the validating ColumnRef
        from .column_ref import ColumnRef

        self.table_columns: list[ColumnSlotIdx] = []
        self.input_exprs = ExprSet()
        validating_colrefs: dict[Expr, Expr] = {}  # key: non-validating colref, value: corresp. validating colref
        for col in columns:
            expr: Expr
            if col.is_computed:
                assert col.value_expr is not None
                # create a copy here so we don't reuse execution state and resolve references to computed columns
                expr = col.value_expr.copy().resolve_computed_cols(resolve_cols=resolve_cols)
                expr = expr.substitute(validating_colrefs)
                expr = self._record_unique_expr(expr, recursive=True)
            else:
                # record a ColumnRef so that references to this column resolve to the same slot idx
                perform_validation = (
                    None
                    if not col.col_type.is_media_type()
                    else col.media_validation == catalog.MediaValidation.ON_WRITE
                )
                expr = ColumnRef(col, perform_validation=perform_validation)
                # recursive=True: needed for validating ColumnRef
                expr = self._record_unique_expr(expr, recursive=True)

                if perform_validation:
                    # if expr is a validating ColumnRef, the input is the non-validating ColumnRef
                    non_validating_colref = expr.components[0]
                    self.input_exprs.add(non_validating_colref)
                    validating_colrefs[non_validating_colref] = expr
                else:
                    self.input_exprs.add(expr)

            self.add_table_column(col, expr.slot_idx)
            self.output_exprs.add(expr)

        # default eval ctx: all output exprs
        self.default_eval_ctx = self.create_eval_ctx(list(self.output_exprs), exclude=unique_input_exprs)

        # references to unstored iterator columns:
        # - those ColumnRefs need to instantiate iterators
        # - we create and record the iterator args here and pass them to their respective ColumnRefs
        # - we do this instead of simply recording the iterator args as a component of those ColumnRefs,
        #   because that would cause them to be evaluated for every single row
        # - the separate eval ctx allows the ColumnRef to materialize the iterator args only when the underlying
        #   iterated object changes
        col_refs = [e for e in self.unique_exprs if isinstance(e, ColumnRef)]

        def refs_unstored_iter_col(col_ref: ColumnRef) -> bool:
            tbl = col_ref.col.tbl
            return (
                tbl.get().is_component_view and tbl.get().is_iterator_column(col_ref.col) and not col_ref.col.is_stored
            )

        unstored_iter_col_refs = [col_ref for col_ref in col_refs if refs_unstored_iter_col(col_ref)]
        component_views = [col_ref.col.tbl for col_ref in unstored_iter_col_refs]
        unstored_iter_args = {view.id: view.get().iterator_args.copy() for view in component_views}
        self.unstored_iter_args = {
            id: self._record_unique_expr(arg, recursive=True) for id, arg in unstored_iter_args.items()
        }

        for col_ref in unstored_iter_col_refs:
            iter_arg_ctx = self.create_eval_ctx([unstored_iter_args[col_ref.col.tbl.id]])
            col_ref.set_iter_arg_ctx(iter_arg_ctx)

        # we guarantee that we can compute the expr DAG in a single front-to-back pass
        for i, expr in enumerate(self.unique_exprs):
            assert expr.slot_idx == i

        # determine transitive dependencies for the purpose of exception propagation
        # (list of set of slot_idxs, indexed by slot_idx)
        # self.dependents = np.zeros((self.num_materialized, self.num_materialized), dtype=bool)
        self.dependencies = np.zeros((self.num_materialized, self.num_materialized), dtype=bool)
        exc_dependencies: list[set[int]] = [set() for _ in range(self.num_materialized)]
        from .column_property_ref import ColumnPropertyRef

        for expr in self.unique_exprs:
            if expr.slot_idx in self.input_expr_slot_idxs:
                # this is input and therefore doesn't depend on other exprs
                continue
            # error properties don't have exceptions themselves
            if isinstance(expr, ColumnPropertyRef) and expr.is_error_prop():
                continue
            dependency_idxs = [d.slot_idx for d in expr.dependencies()]
            self.dependencies[expr.slot_idx, dependency_idxs] = True
            for d in expr.dependencies():
                exc_dependencies[expr.slot_idx].add(d.slot_idx)
                exc_dependencies[expr.slot_idx].update(exc_dependencies[d.slot_idx])

        self.num_dependencies = np.sum(self.dependencies, axis=1)
        self.dependents = self.dependencies.T
        self.transitive_dependents = np.zeros((self.num_materialized, self.num_materialized), dtype=bool)
        for i in reversed(range(self.num_materialized)):
            self.transitive_dependents[i] = self.dependents[i] | np.any(
                self.transitive_dependents[self.dependents[i]], axis=0
            )

        self._exc_dependents = [set() for _ in range(self.num_materialized)]
        for expr in self.unique_exprs:
            assert expr.slot_idx is not None
            for d_idx in exc_dependencies[expr.slot_idx]:
                self._exc_dependents[d_idx].add(expr.slot_idx)

        self.output_expr_ids = [set() for _ in range(self.num_materialized)]
        for e in self.output_exprs:
            self._record_output_expr_id(e, e.slot_idx)

    def add_table_column(self, col: catalog.Column, slot_idx: int) -> None:
        """Record a column that is part of the table row"""
        self.table_columns.append(ColumnSlotIdx(col, slot_idx))

    def output_slot_idxs(self) -> list[ColumnSlotIdx]:
        """Return ColumnSlotIdx for output columns"""
        return self.table_columns

    @property
    def num_materialized(self) -> int:
        return self.next_slot_idx

    def get_output_exprs(self) -> list[Expr]:
        """Returns exprs that were requested in the c'tor and require evaluation"""
        return list(self.output_exprs)

    def _next_slot_idx(self) -> int:
        result = self.next_slot_idx
        self.next_slot_idx += 1
        return result

    def _record_unique_expr(self, expr: Expr, recursive: bool) -> Expr:
        """Records the expr if it's not a duplicate and assigns a slot idx to expr and its components"
        Returns:
            the unique expr
        """
        if expr in self.unique_exprs:
            # expr is a duplicate: we use the original instead
            return self.unique_exprs[expr]

        # expr value needs to be computed via Expr.eval()
        if recursive:
            for i, c in enumerate(expr.components):
                # make sure we only refer to components that have themselves been recorded
                expr.components[i] = self._record_unique_expr(c, True)
        assert expr.slot_idx is None
        expr.slot_idx = self._next_slot_idx()
        self.unique_exprs.add(expr)
        return expr

    def _record_output_expr_id(self, e: Expr, output_expr_id: int) -> None:
        assert e.slot_idx is not None
        assert output_expr_id is not None
        if e.slot_idx in self.input_expr_slot_idxs:
            return
        self.output_expr_ids[e.slot_idx].add(output_expr_id)
        for d in e.dependencies():
            self._record_output_expr_id(d, output_expr_id)

    def _compute_dependencies(
        self, target_slot_idxs: list[int], excluded_slot_idxs: list[int], target_scope: Optional[ExprScope] = None
    ) -> list[int]:
        """Compute exprs needed to materialize the given target slots, excluding 'excluded_slot_idxs'

        If target_scope != None, stops transitive dependency resolution when leaving target_scope (ie, includes
        immediate dependents that aren't in target_scope, but doesn't resolve those).
        """
        dependencies: list[set[int]] = [set() for _ in range(self.num_materialized)]  # indexed by slot_idx
        # doing this front-to-back ensures that we capture transitive dependencies
        max_target_slot_idx = max(target_slot_idxs)
        for expr in self.unique_exprs:
            if expr.slot_idx > max_target_slot_idx:
                # we're done
                break
            if expr.slot_idx in excluded_slot_idxs:
                continue
            if expr.slot_idx in self.input_expr_slot_idxs:
                # this is input and therefore doesn't depend on other exprs
                continue
            if target_scope is not None and expr.scope() != target_scope:
                # don't resolve dependencies outside of target_scope
                continue
            for d in expr.dependencies():
                assert d.slot_idx is not None, f'{expr}, {d}'
                if d.slot_idx in excluded_slot_idxs:
                    continue
                dependencies[expr.slot_idx].add(d.slot_idx)
                dependencies[expr.slot_idx].update(dependencies[d.slot_idx])
        # merge dependencies and convert to list
        return sorted(set().union(*[dependencies[i] for i in target_slot_idxs]))

    def set_slot_idxs(self, expr_list: Sequence[Expr], remove_duplicates: bool = True) -> None:
        """
        Recursively sets slot_idx in expr_list and its components

        remove_duplicates == True: removes duplicates in-place
        """
        for e in expr_list:
            self.__set_slot_idxs_aux(e)
        if remove_duplicates:
            # only allowed if `expr_list` is a mutable list
            assert isinstance(expr_list, list)
            deduped = list(ExprSet(expr_list))
            expr_list[:] = deduped

    def __set_slot_idxs_aux(self, e: Expr) -> None:
        """Recursively sets slot_idx in e and its components"""
        if e not in self.unique_exprs:
            return
        e.slot_idx = self.unique_exprs[e].slot_idx
        for c in e.components:
            self.__set_slot_idxs_aux(c)

    def get_dependencies(
        self, targets: Iterable[Expr], exclude: Optional[Iterable[Expr]] = None, limit_scope: bool = True
    ) -> list[Expr]:
        """
        Return list of dependencies needed to evaluate the given target exprs (expressed as slot idxs).
        The exprs given in 'exclude' are excluded.
        If limit_scope == True, only returns dependencies in the same scope and immediate (ie, not transitive)
        dependencies from enclosing scopes.

        Returns:
            list of Exprs from unique_exprs (= with slot_idx set)
        """
        targets = list(targets)
        if exclude is None:
            exclude = []
        if len(targets) == 0:
            return []
        # make sure we only refer to recorded exprs
        targets = [self.unique_exprs[e] for e in targets]
        target_scope: Optional[ExprScope] = None
        if limit_scope:
            # make sure all targets are from the same scope
            target_scopes = {e.scope() for e in targets}
            assert len(target_scopes) == 1
            target_scope = target_scopes.pop()
        exclude = [self.unique_exprs[e] for e in exclude]
        target_slot_idxs = [e.slot_idx for e in targets]
        excluded_slot_idxs = [e.slot_idx for e in exclude]
        all_dependencies = set(
            self._compute_dependencies(target_slot_idxs, excluded_slot_idxs, target_scope=target_scope)
        )
        all_dependencies.update(target_slot_idxs)
        result_ids = list(all_dependencies)
        result_ids.sort()
        return [self.unique_exprs[id] for id in result_ids]

    def create_eval_ctx(
        self, targets: Iterable[Expr], exclude: Optional[Iterable[Expr]] = None, limit_scope: bool = True
    ) -> EvalCtx:
        """Return EvalCtx for targets"""
        targets = list(targets)
        if exclude is None:
            exclude = []
        if len(targets) == 0:
            return self.EvalCtx([], [], [], [])
        dependencies = self.get_dependencies(targets, exclude, limit_scope=limit_scope)
        targets = [self.unique_exprs[e] for e in targets]
        target_slot_idxs = [e.slot_idx for e in targets]
        ctx_slot_idxs = [e.slot_idx for e in dependencies]
        return self.EvalCtx(
            slot_idxs=ctx_slot_idxs,
            exprs=[self.unique_exprs[slot_idx] for slot_idx in ctx_slot_idxs],
            target_slot_idxs=target_slot_idxs,
            target_exprs=targets,
        )

    def set_exc(self, data_row: DataRow, slot_idx: int, exc: Exception) -> None:
        """Record an exception in data_row and propagate it to dependents"""
        data_row.set_exc(slot_idx, exc)
        for idx in self._exc_dependents[slot_idx]:
            data_row.set_exc(idx, exc)

    def eval(
        self,
        data_row: DataRow,
        ctx: EvalCtx,
        profile: Optional[ExecProfile] = None,
        ignore_errors: bool = False,
        force_eval: Optional[ExprScope] = None,
    ) -> None:
        """
        Populates the slots in data_row given in ctx.
        If an expr.eval() raises an exception, records the exception in the corresponding slot of data_row
        and omits any of that expr's dependents's eval().
        profile: if present, populated with execution time of each expr.eval() call; indexed by expr.slot_idx
        ignore_errors: if False, raises ExprEvalError if any expr.eval() raises an exception
        force_eval: forces exprs in the specified scope to be reevaluated, even if they already have a value
        """
        for expr in ctx.exprs:
            assert expr.slot_idx >= 0
            if expr.scope() != force_eval and (data_row.has_val[expr.slot_idx] or data_row.has_exc(expr.slot_idx)):
                continue
            try:
                start_time = time.perf_counter()
                expr.eval(data_row, self)
                if profile is not None:
                    profile.eval_time[expr.slot_idx] += time.perf_counter() - start_time
                    profile.eval_count[expr.slot_idx] += 1
            except Exception as exc:
                _, _, exc_tb = sys.exc_info()
                self.set_exc(data_row, expr.slot_idx, exc)
                if not ignore_errors:
                    input_vals = [data_row[d.slot_idx] for d in expr.dependencies()]
                    raise excs.ExprEvalError(
                        expr, f'expression {expr}', data_row.get_exc(expr.slot_idx), exc_tb, input_vals, 0
                    ) from exc

    def create_table_row(self, data_row: DataRow, exc_col_ids: set[int]) -> tuple[dict[str, Any], int]:
        """Create a table row from the slots that have an output column assigned

        Return tuple[dict that represents a stored row (can be passed to sql.insert()), # of exceptions]
            This excludes system columns.
        """
        num_excs = 0
        table_row: dict[str, Any] = {}
        for info in self.table_columns:
            col, slot_idx = info.col, info.slot_idx
            if data_row.has_exc(slot_idx):
                # exceptions get stored in the errortype/-msg columns
                exc = data_row.get_exc(slot_idx)
                num_excs += 1
                exc_col_ids.add(col.id)
                table_row[col.store_name()] = None
                table_row[col.errortype_store_name()] = type(exc).__name__
                table_row[col.errormsg_store_name()] = str(exc)
            else:
                if col.col_type.is_image_type() and data_row.file_urls[slot_idx] is None:
                    # we have yet to store this image
                    filepath = str(MediaStore.prepare_media_path(col.tbl.id, col.id, col.tbl.get().version))
                    data_row.flush_img(slot_idx, filepath)
                val = data_row.get_stored_val(slot_idx, col.sa_col.type)
                table_row[col.store_name()] = val
                # we unfortunately need to set these, even if there are no errors
                table_row[col.errortype_store_name()] = None
                table_row[col.errormsg_store_name()] = None

        return table_row, num_excs



"""
Pixeltable [UDFs](https://pixeltable.readme.io/docs/user-defined-functions-udfs)
that wrap various endpoints from the Together AI API. In order to use them, you must
first `pip install together` and configure your Together AI credentials, as described in
the [Working with Together AI](https://pixeltable.readme.io/docs/together-ai) tutorial.
"""

import base64
import io
from typing import TYPE_CHECKING, Callable, Optional, TypeVar

import numpy as np
import PIL.Image
import requests
import tenacity

import pixeltable as pxt
import pixeltable.exceptions as excs
from pixeltable import env
from pixeltable.func import Batch
from pixeltable.utils.code import local_public_names

if TYPE_CHECKING:
    import together  # type: ignore[import-untyped]


@env.register_client('together')
def _(api_key: str) -> 'together.AsyncTogether':
    import together

    return together.AsyncTogether(api_key=api_key)


def _together_client() -> 'together.AsyncTogether':
    return env.Env.get().get_client('together')


T = TypeVar('T')


def _retry(fn: Callable[..., T]) -> Callable[..., T]:
    import together

    return tenacity.retry(
        retry=tenacity.retry_if_exception_type(together.error.RateLimitError),
        wait=tenacity.wait_random_exponential(multiplier=1, max=60),
        stop=tenacity.stop_after_attempt(20),
    )(fn)


@pxt.udf(resource_pool='request-rate:together:chat')
async def completions(
    prompt: str,
    *,
    model: str,
    max_tokens: Optional[int] = None,
    stop: Optional[list] = None,
    temperature: Optional[float] = None,
    top_p: Optional[float] = None,
    top_k: Optional[int] = None,
    repetition_penalty: Optional[float] = None,
    logprobs: Optional[int] = None,
    echo: Optional[bool] = None,
    n: Optional[int] = None,
    safety_model: Optional[str] = None,
) -> dict:
    """
    Generate completions based on a given prompt using a specified model.

    Equivalent to the Together AI `completions` API endpoint.
    For additional details, see: <https://docs.together.ai/reference/completions-1>

    Request throttling:
    Applies the rate limit set in the config (section `together.rate_limits`, key `chat`). If no rate
    limit is configured, uses a default of 600 RPM.

    __Requirements:__

    - `pip install together`

    Args:
        prompt: A string providing context for the model to complete.
        model: The name of the model to query.

    For details on the other parameters, see: <https://docs.together.ai/reference/completions-1>

    Returns:
        A dictionary containing the response and other metadata.

    Examples:
        Add a computed column that applies the model `mistralai/Mixtral-8x7B-v0.1` to an existing Pixeltable column
        `tbl.prompt` of the table `tbl`:

        >>> tbl.add_computed_column(response=completions(tbl.prompt, model='mistralai/Mixtral-8x7B-v0.1'))
    """
    result = await _together_client().completions.create(
        prompt=prompt,
        model=model,
        max_tokens=max_tokens,
        stop=stop,
        temperature=temperature,
        top_p=top_p,
        top_k=top_k,
        repetition_penalty=repetition_penalty,
        logprobs=logprobs,
        echo=echo,
        n=n,
        safety_model=safety_model,
    )
    return result.dict()


@pxt.udf(resource_pool='request-rate:together:chat')
async def chat_completions(
    messages: list[dict[str, str]],
    *,
    model: str,
    max_tokens: Optional[int] = None,
    stop: Optional[list[str]] = None,
    temperature: Optional[float] = None,
    top_p: Optional[float] = None,
    top_k: Optional[int] = None,
    repetition_penalty: Optional[float] = None,
    logprobs: Optional[int] = None,
    echo: Optional[bool] = None,
    n: Optional[int] = None,
    safety_model: Optional[str] = None,
    response_format: Optional[dict] = None,
    tools: Optional[dict] = None,
    tool_choice: Optional[dict] = None,
) -> dict:
    """
    Generate chat completions based on a given prompt using a specified model.

    Equivalent to the Together AI `chat/completions` API endpoint.
    For additional details, see: <https://docs.together.ai/reference/chat-completions-1>

    Request throttling:
    Applies the rate limit set in the config (section `together.rate_limits`, key `chat`). If no rate
    limit is configured, uses a default of 600 RPM.

    __Requirements:__

    - `pip install together`

    Args:
        messages: A list of messages comprising the conversation so far.
        model: The name of the model to query.

    For details on the other parameters, see: <https://docs.together.ai/reference/chat-completions-1>

    Returns:
        A dictionary containing the response and other metadata.

    Examples:
        Add a computed column that applies the model `mistralai/Mixtral-8x7B-v0.1` to an existing Pixeltable column
        `tbl.prompt` of the table `tbl`:

        >>> messages = [{'role': 'user', 'content': tbl.prompt}]
        ... tbl.add_computed_column(response=chat_completions(messages, model='mistralai/Mixtral-8x7B-v0.1'))
    """
    result = await _together_client().chat.completions.create(
        messages=messages,
        model=model,
        max_tokens=max_tokens,
        stop=stop,
        temperature=temperature,
        top_p=top_p,
        top_k=top_k,
        repetition_penalty=repetition_penalty,
        logprobs=logprobs,
        echo=echo,
        n=n,
        safety_model=safety_model,
        response_format=response_format,
        tools=tools,
        tool_choice=tool_choice,
    )
    return result.dict()


_embedding_dimensions_cache = {
    'togethercomputer/m2-bert-80M-2k-retrieval': 768,
    'togethercomputer/m2-bert-80M-8k-retrieval': 768,
    'togethercomputer/m2-bert-80M-32k-retrieval': 768,
    'WhereIsAI/UAE-Large-V1': 1024,
    'BAAI/bge-large-en-v1.5': 1024,
    'BAAI/bge-base-en-v1.5': 768,
    'sentence-transformers/msmarco-bert-base-dot-v5': 768,
    'bert-base-uncased': 768,
}


@pxt.udf(batch_size=32, resource_pool='request-rate:together:embeddings')
async def embeddings(input: Batch[str], *, model: str) -> Batch[pxt.Array[(None,), pxt.Float]]:
    """
    Query an embedding model for a given string of text.

    Equivalent to the Together AI `embeddings` API endpoint.
    For additional details, see: <https://docs.together.ai/reference/embeddings-2>

    Request throttling:
    Applies the rate limit set in the config (section `together.rate_limits`, key `embeddings`). If no rate
    limit is configured, uses a default of 600 RPM.

    __Requirements:__

    - `pip install together`

    Args:
        input: A string providing the text for the model to embed.
        model: The name of the embedding model to use.

    Returns:
        An array representing the application of the given embedding to `input`.

    Examples:
        Add a computed column that applies the model `togethercomputer/m2-bert-80M-8k-retrieval`
        to an existing Pixeltable column `tbl.text` of the table `tbl`:

        >>> tbl.add_computed_column(response=embeddings(tbl.text, model='togethercomputer/m2-bert-80M-8k-retrieval'))
    """
    result = await _together_client().embeddings.create(input=input, model=model)
    return [np.array(data.embedding, dtype=np.float64) for data in result.data]


@embeddings.conditional_return_type
def _(model: str) -> pxt.ArrayType:
    if model not in _embedding_dimensions_cache:
        # TODO: find some other way to retrieve a sample
        return pxt.ArrayType((None,), dtype=pxt.FloatType())
    dimensions = _embedding_dimensions_cache[model]
    return pxt.ArrayType((dimensions,), dtype=pxt.FloatType())


@pxt.udf(resource_pool='request-rate:together:images')
async def image_generations(
    prompt: str,
    *,
    model: str,
    steps: Optional[int] = None,
    seed: Optional[int] = None,
    height: Optional[int] = None,
    width: Optional[int] = None,
    negative_prompt: Optional[str] = None,
) -> PIL.Image.Image:
    """
    Generate images based on a given prompt using a specified model.

    Equivalent to the Together AI `images/generations` API endpoint.
    For additional details, see: <https://docs.together.ai/reference/post_images-generations>

    Request throttling:
    Applies the rate limit set in the config (section `together.rate_limits`, key `images`). If no rate
    limit is configured, uses a default of 600 RPM.

    __Requirements:__

    - `pip install together`

    Args:
        prompt: A description of the desired images.
        model: The model to use for image generation.

    For details on the other parameters, see: <https://docs.together.ai/reference/post_images-generations>

    Returns:
        The generated image.

    Examples:
        Add a computed column that applies the model `stabilityai/stable-diffusion-xl-base-1.0`
        to an existing Pixeltable column `tbl.prompt` of the table `tbl`:

        >>> tbl.add_computed_column(
        ...     response=image_generations(tbl.prompt, model='stabilityai/stable-diffusion-xl-base-1.0')
        ... )
    """
    result = await _together_client().images.generate(
        prompt=prompt, model=model, steps=steps, seed=seed, height=height, width=width, negative_prompt=negative_prompt
    )
    if result.data[0].b64_json is not None:
        b64_bytes = base64.b64decode(result.data[0].b64_json)
        img = PIL.Image.open(io.BytesIO(b64_bytes))
        img.load()
        return img
    if result.data[0].url is not None:
        try:
            resp = requests.get(result.data[0].url)
            with io.BytesIO(resp.content) as fp:
                image = PIL.Image.open(fp)
                image.load()
                return image
        except Exception as exc:
            raise excs.Error('Failed to download generated image from together.ai.') from exc
    raise excs.Error('Response does not contain a generated image.')


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



import builtins
import typing
from typing import Any, Callable, Optional, Union

import sqlalchemy as sql

from pixeltable import exceptions as excs, exprs, func, type_system as ts
from pixeltable.utils.code import local_public_names

from typing import _GenericAlias  # type: ignore[attr-defined]  # isort: skip


# TODO: remove and replace calls with astype()
def cast(expr: exprs.Expr, target_type: Union[ts.ColumnType, type, _GenericAlias]) -> exprs.Expr:
    expr.col_type = ts.ColumnType.normalize_type(target_type)
    return expr


T = typing.TypeVar('T')


@func.uda(allows_window=True, type_substitutions=({T: Optional[int]}, {T: Optional[float]}))  # type: ignore[misc]
class sum(func.Aggregator, typing.Generic[T]):
    """Sums the selected integers or floats."""

    def __init__(self) -> None:
        self.sum: T = None

    def update(self, val: T) -> None:
        if val is None:
            return
        if self.sum is None:
            self.sum = val
        else:
            self.sum += val  # type: ignore[operator]

    def value(self) -> T:
        return self.sum


@sum.to_sql
def _(val: sql.ColumnElement) -> Optional[sql.ColumnElement]:
    # This can produce a Decimal. We are deliberately avoiding an explicit cast to a Bigint here, because that can
    # cause overflows in Postgres. We're instead doing the conversion to the target type in SqlNode.__iter__().
    return sql.sql.func.sum(val)


@func.uda(
    allows_window=True,
    # Allow counting non-null values of any type
    # TODO: should we have an "Any" type that can be used here?
    type_substitutions=tuple(
        {T: Optional[t]}  # type: ignore[misc]
        for t in (
            ts.String,
            ts.Int,
            ts.Float,
            ts.Bool,
            ts.Timestamp,
            ts.Array,
            ts.Json,
            ts.Image,
            ts.Video,
            ts.Audio,
            ts.Document,
        )
    ),
)
class count(func.Aggregator, typing.Generic[T]):
    def __init__(self) -> None:
        self.count = 0

    def update(self, val: T) -> None:
        if val is not None:
            self.count += 1

    def value(self) -> int:
        return self.count


@count.to_sql
def _(val: sql.ColumnElement) -> Optional[sql.ColumnElement]:
    return sql.sql.func.count(val)


@func.uda(
    allows_window=True,
    type_substitutions=tuple({T: Optional[t]} for t in (str, int, float, bool, ts.Timestamp)),  # type: ignore[misc]
)
class min(func.Aggregator, typing.Generic[T]):
    def __init__(self) -> None:
        self.val: T = None

    def update(self, val: T) -> None:
        if val is None:
            return
        if self.val is None:
            self.val = val
        else:
            self.val = builtins.min(self.val, val)  # type: ignore[call-overload]

    def value(self) -> T:
        return self.val


@min.to_sql
def _(val: sql.ColumnElement) -> Optional[sql.ColumnElement]:
    if val.type.python_type is bool:
        # TODO: min/max aggregation of booleans is not supported in Postgres (but it is in Python).
        # Right now we simply force the computation to be done in Python; we might consider implementing an alternate
        # way of doing it in SQL. (min/max of booleans is simply logical and/or, respectively.)
        return None
    return sql.sql.func.min(val)


@func.uda(
    allows_window=True,
    type_substitutions=tuple({T: Optional[t]} for t in (str, int, float, bool, ts.Timestamp)),  # type: ignore[misc]
)
class max(func.Aggregator, typing.Generic[T]):
    def __init__(self) -> None:
        self.val: T = None

    def update(self, val: T) -> None:
        if val is None:
            return
        if self.val is None:
            self.val = val
        else:
            self.val = builtins.max(self.val, val)  # type: ignore[call-overload]

    def value(self) -> T:
        return self.val


@max.to_sql
def _(val: sql.ColumnElement) -> Optional[sql.ColumnElement]:
    if val.type.python_type is bool:
        # TODO: see comment in @min.to_sql.
        return None
    return sql.sql.func.max(val)


@func.uda(type_substitutions=({T: Optional[int]}, {T: Optional[float]}))  # type: ignore[misc]
class mean(func.Aggregator, typing.Generic[T]):
    def __init__(self) -> None:
        self.sum: T = None
        self.count = 0

    def update(self, val: T) -> None:
        if val is None:
            return
        if self.sum is None:
            self.sum = val
        else:
            self.sum += val  # type: ignore[operator]
        self.count += 1

    def value(self) -> Optional[float]:  # Always a float
        if self.count == 0:
            return None
        return self.sum / self.count  # type: ignore[operator]


@mean.to_sql
def _(val: sql.ColumnElement) -> Optional[sql.ColumnElement]:
    return sql.sql.func.avg(val)


def map(expr: exprs.Expr, fn: Callable[[exprs.Expr], Any]) -> exprs.Expr:
    target_expr: exprs.Expr
    try:
        target_expr = exprs.Expr.from_object(fn(exprs.json_path.RELATIVE_PATH_ROOT))
    except Exception as e:
        raise excs.Error(
            'Failed to evaluate map function. '
            '(The `fn` argument to `map()` must produce a valid Pixeltable expression.)'
        ) from e
    return exprs.JsonMapper(expr, target_expr)


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



import os
from typing import Any, Optional, Union

import numpy as np
import pandas as pd
from pandas._typing import DtypeObj  # For pandas dtype type hints
from pandas.api.types import is_datetime64_any_dtype, is_extension_array_dtype

import pixeltable as pxt
import pixeltable.exceptions as excs


def import_pandas(
    tbl_name: str,
    df: pd.DataFrame,
    *,
    schema_overrides: Optional[dict[str, Any]] = None,
    primary_key: Optional[Union[str, list[str]]] = None,
    num_retained_versions: int = 10,
    comment: str = '',
) -> pxt.Table:
    """Creates a new base table from a Pandas
    [`DataFrame`](https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.html), with the
    specified name. The schema of the table will be inferred from the DataFrame.

    The column names of the new table will be identical to those in the DataFrame, as long as they are valid
    Pixeltable identifiers. If a column name is not a valid Pixeltable identifier, it will be normalized according to
    the following procedure:
    - first replace any non-alphanumeric characters with underscores;
    - then, preface the result with the letter 'c' if it begins with a number or an underscore;
    - then, if there are any duplicate column names, suffix the duplicates with '_2', '_3', etc., in column order.

    Args:
        tbl_name: The name of the table to create.
        df: The Pandas `DataFrame`.
        schema_overrides: If specified, then for each (name, type) pair in `schema_overrides`, the column with
            name `name` will be given type `type`, instead of being inferred from the `DataFrame`. The keys in
            `schema_overrides` should be the column names of the `DataFrame` (whether or not they are valid
            Pixeltable identifiers).

    Returns:
        A handle to the newly created [`Table`][pixeltable.Table].
    """
    return pxt.create_table(
        tbl_name,
        source=df,
        schema_overrides=schema_overrides,
        primary_key=primary_key,
        num_retained_versions=num_retained_versions,
        comment=comment,
    )


def import_csv(
    tbl_name: str,
    filepath_or_buffer: Union[str, os.PathLike],
    schema_overrides: Optional[dict[str, Any]] = None,
    primary_key: Optional[Union[str, list[str]]] = None,
    num_retained_versions: int = 10,
    comment: str = '',
    **kwargs: Any,
) -> pxt.Table:
    """
    Creates a new base table from a csv file. This is a convenience method and is equivalent
    to calling `import_pandas(table_path, pd.read_csv(filepath_or_buffer, **kwargs), schema=schema)`.
    See the Pandas documentation for [`read_csv`](https://pandas.pydata.org/docs/reference/api/pandas.read_csv.html)
    for more details.

    Returns:
        A handle to the newly created [`Table`][pixeltable.Table].
    """
    return pxt.create_table(
        tbl_name,
        source=filepath_or_buffer,
        schema_overrides=schema_overrides,
        primary_key=primary_key,
        num_retained_versions=num_retained_versions,
        comment=comment,
        extra_args=kwargs,
    )


def import_excel(
    tbl_name: str,
    io: Union[str, os.PathLike],
    *,
    schema_overrides: Optional[dict[str, Any]] = None,
    primary_key: Optional[Union[str, list[str]]] = None,
    num_retained_versions: int = 10,
    comment: str = '',
    **kwargs: Any,
) -> pxt.Table:
    """
    Creates a new base table from an Excel (.xlsx) file. This is a convenience method and is
    equivalent to calling `import_pandas(table_path, pd.read_excel(io, *args, **kwargs), schema=schema)`.
    See the Pandas documentation for [`read_excel`](https://pandas.pydata.org/docs/reference/api/pandas.read_excel.html)
    for more details.

    Returns:
        A handle to the newly created [`Table`][pixeltable.Table].
    """
    return pxt.create_table(
        tbl_name,
        source=io,
        schema_overrides=schema_overrides,
        primary_key=primary_key,
        num_retained_versions=num_retained_versions,
        comment=comment,
        extra_args=kwargs,
    )


def _df_check_primary_key_values(df: pd.DataFrame, primary_key: list[str]) -> None:
    for pd_name in primary_key:
        # This can be faster for large DataFrames
        has_nulls = df[pd_name].count() < len(df)
        if has_nulls:
            raise excs.Error(f'Primary key column `{pd_name}` cannot contain null values.')


def df_infer_schema(
    df: pd.DataFrame, schema_overrides: dict[str, pxt.ColumnType], primary_key: list[str]
) -> dict[str, pxt.ColumnType]:
    """
    Infers a Pixeltable schema from a Pandas DataFrame.

    Returns:
        A tuple containing a Pixeltable schema and a list of primary key column names.
    """
    pd_schema: dict[str, pxt.ColumnType] = {}
    for pd_name, pd_dtype in zip(df.columns, df.dtypes):
        if pd_name in schema_overrides:
            pxt_type = schema_overrides[pd_name]
        else:
            pxt_type = __pd_coltype_to_pxt_type(pd_dtype, df[pd_name], pd_name not in primary_key)
        pd_schema[pd_name] = pxt_type

    return pd_schema


def __pd_dtype_to_pxt_type(pd_dtype: DtypeObj, nullable: bool) -> Optional[pxt.ColumnType]:
    """
    Determines a pixeltable ColumnType from a pandas dtype

    Args:
        pd_dtype: A pandas dtype object

    Returns:
        pxt.ColumnType: A pixeltable ColumnType
    """
    # Pandas extension arrays / types (Int64, boolean, string[pyarrow], etc.) are not directly
    # compatible with NumPy dtypes
    # The timezone-aware datetime64[ns, tz=] dtype is a pandas extension dtype
    if is_datetime64_any_dtype(pd_dtype):
        return pxt.TimestampType(nullable=nullable)
    if is_extension_array_dtype(pd_dtype):
        return None
    # Most other pandas dtypes are directly NumPy compatible
    assert isinstance(pd_dtype, np.dtype)
    return pxt.ArrayType.from_np_dtype(pd_dtype, nullable)


def __pd_coltype_to_pxt_type(pd_dtype: DtypeObj, data_col: pd.Series, nullable: bool) -> pxt.ColumnType:
    """
    Infers a Pixeltable type based on a pandas dtype.
    """
    pxttype = __pd_dtype_to_pxt_type(pd_dtype, nullable)
    if pxttype is not None:
        return pxttype

    if pd_dtype == np.object_:
        # The `object_` dtype can mean all sorts of things; see if we can infer the Pixeltable type
        # based on the actual data in `data_col`.
        # First drop any null values (they don't contribute to type inference).
        data_col = data_col.dropna()

        if len(data_col) == 0:
            # No non-null values; default to FloatType (the Pandas type of an all-NaN column)
            return pxt.FloatType(nullable=nullable)

        inferred_type = pxt.ColumnType.infer_common_literal_type(data_col)
        if inferred_type is None:
            # Fallback on StringType if everything else fails
            return pxt.StringType(nullable=nullable)
        else:
            return inferred_type.copy(nullable=nullable)

    raise excs.Error(f'Could not infer Pixeltable type of column: {data_col.name} (dtype: {pd_dtype})')


def _df_row_to_pxt_row(
    row: tuple[Any, ...], schema: dict[str, pxt.ColumnType], col_mapping: Optional[dict[str, str]]
) -> dict[str, Any]:
    """Convert a row to insertable format"""
    pxt_row: dict[str, Any] = {}
    for val, (col_name, pxt_type) in zip(row[1:], schema.items()):
        pxt_name = col_mapping.get(col_name, col_name)
        nval: Any
        if pxt_type.is_float_type():
            nval = float(val)
        elif isinstance(val, float) and np.isnan(val):
            # pandas uses NaN for empty cells, even for types other than float;
            # for any type but a float, convert these to None
            nval = None
        elif pxt_type.is_int_type():
            nval = int(val)
        elif pxt_type.is_bool_type():
            nval = bool(val)
        elif pxt_type.is_string_type():
            nval = str(val)
        elif pxt_type.is_timestamp_type():
            if pd.isnull(val):
                # pandas has the bespoke 'NaT' type for a missing timestamp; postgres is very
                # much not-ok with it. (But if we convert it to None and then load out the
                # table contents as a pandas DataFrame, it will correctly restore the 'NaT'!)
                nval = None
            else:
                nval = pd.Timestamp(val).to_pydatetime()
        else:
            nval = val
        pxt_row[pxt_name] = nval
    return pxt_row



import asyncio
import threading
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Coroutine, TypeVar

T = TypeVar('T')


# TODO This is a temporary hack to be able to run async UDFs in contexts that are not properly handled by the existing
#   scheduler logic (e.g., as an embedding function as part of a similarity lookup). Once the scheduler is fully
#   general, it can be removed.


def run_coroutine_synchronously(coroutine: Coroutine[Any, Any, T], timeout: float = 30) -> T:
    """
    Runs the given coroutine synchronously, even if called in the context of a running event loop.
    """

    def run_in_new_loop() -> T:
        new_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(new_loop)
        try:
            return new_loop.run_until_complete(coroutine)
        finally:
            new_loop.close()

    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        # No event loop; just call `asyncio.run()`
        return asyncio.run(coroutine)

    if threading.current_thread() is threading.main_thread():
        if not loop.is_running():
            return loop.run_until_complete(coroutine)
        else:
            with ThreadPoolExecutor() as pool:
                future = pool.submit(run_in_new_loop)
                return future.result(timeout=timeout)
    else:
        return asyncio.run_coroutine_threadsafe(coroutine, loop).result()



import inspect
import logging
from typing import Any

import cloudpickle  # type: ignore[import-untyped]
import sqlalchemy as sql

from pixeltable.metadata import register_converter
from pixeltable.metadata.schema import Function

_logger = logging.getLogger('pixeltable')


@register_converter(version=15)
def _(engine: sql.engine.Engine) -> None:
    with engine.begin() as conn:
        for row in conn.execute(sql.select(Function)):
            id, _, md, binary_obj = row
            md['md'] = __update_md(md['md'], binary_obj)
            _logger.info(f'Updating function: {id}')
            conn.execute(sql.update(Function).where(Function.id == id).values(md=md))


def __update_md(orig_d: dict, binary_obj: bytes) -> Any:
    # construct dict produced by CallableFunction.to_store()
    py_fn = cloudpickle.loads(binary_obj)
    py_params = inspect.signature(py_fn).parameters
    return_type = orig_d['return_type']
    params: list[dict] = []
    for name, col_type_dict, kind_int, is_batched in orig_d['parameters']:
        default = py_params[name].default
        kind = inspect._ParameterKind(kind_int)
        params.append(
            {
                'name': name,
                'col_type': col_type_dict,
                'kind': str(kind),
                'is_batched': is_batched,
                'has_default': default is not inspect.Parameter.empty,
                'default': None if default is inspect.Parameter.empty else default,
            }
        )
    is_batched = 'batch_size' in orig_d
    d = {
        'signature': {'return_type': return_type, 'parameters': params, 'is_batched': is_batched},
        'batch_size': orig_d['batch_size'] if is_batched else None,
    }
    return d



from typing import Any, Optional

import sqlalchemy as sql

from pixeltable.metadata import register_converter
from pixeltable.metadata.converters.util import convert_table_md


@register_converter(version=24)
def _(engine: sql.engine.Engine) -> None:
    convert_table_md(engine, substitution_fn=__substitute_md)


def __substitute_md(k: Optional[str], v: Any) -> Optional[tuple[Optional[str], Any]]:
    from pixeltable import func
    from pixeltable.func.globals import resolve_symbol

    if (
        isinstance(v, dict)
        and '_classpath' in v
        and v['_classpath']
        in (
            'pixeltable.func.callable_function.CallableFunction',
            'pixeltable.func.aggregate_function.AggregateFunction',
            'pixeltable.func.expr_template_function.ExprTemplateFunction',
        )
    ):
        if 'path' in v:
            assert 'signature' not in v
            f = resolve_symbol(__substitute_path(v['path']))
            assert isinstance(f, func.Function)
            v['signature'] = f.signatures[0].as_dict()
        return k, v

    if isinstance(v, dict) and '_classname' in v and v['_classname'] == 'FunctionCall':
        # Correct an older serialization mechanism where Expr elements of FunctionCall args and
        # kwargs were indicated with idx == -1 rather than None. This was fixed for InlineList
        # and InlineDict back in convert_20, but not for FunctionCall.
        assert 'args' in v and isinstance(v['args'], list)
        assert 'kwargs' in v and isinstance(v['kwargs'], dict)
        v['args'] = [(None, arg) if idx == -1 else (idx, arg) for idx, arg in v['args']]
        v['kwargs'] = {k: (None, arg) if idx == -1 else (idx, arg) for k, (idx, arg) in v['kwargs'].items()}
        return k, v

    return None


def __substitute_path(path: str) -> str:
    # Starting with version 25, function signatures are preserved in metadata. To migrate from older
    # versions, it's necessary to resolve the function symbol to get the signature. The following
    # adjustment is necessary for function names that are stored in db artifacts of version < 25, but
    # have changed in some version > 25.
    if path in ('pixeltable.functions.huggingface.clip_text', 'pixeltable.functions.huggingface.clip_image'):
        return 'pixeltable.functions.huggingface.clip'
    return path
